import os

import numpy as np
import pandas as pd

from lingam.rcd import RCD

def test_fit_success():
    # causal direction: x5 --> x0, x3 --> x1 --> x2, x4 <-- x6
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x5 = get_external_effect(n_samples)
    x6 = get_external_effect(n_samples)
    x0 = 0.6*x5 + get_external_effect(n_samples)
    x3 = 0.5*x5 + get_external_effect(n_samples)
    x1 = 1.0*x0 + 1.0*x3 + get_external_effect(n_samples)
    x2 = 0.8*x1 - 0.6*x6 + get_external_effect(n_samples)
    x4 = 1.0*x1 - 0.5*x6 + get_external_effect(n_samples)
    # x5 and x6 are latent confounders
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T,
                     columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    # default
    model = RCD()
    model.fit(X)

    ans = model.ancestors_list_
    am = model.adjacency_matrix_
    p_values = model.get_error_independence_p_values(X)

    # max_explanatory_num=3
    model = RCD(max_explanatory_num=3)
    model.fit(X)

    # max_explanatory_num=1
    model = RCD(max_explanatory_num=1)
    model.fit(X)

    # cor_alpha=0.1
    model = RCD(cor_alpha=0.1)
    model.fit(X)

    # ind_alpha=0.1
    model = RCD(ind_alpha=0.1)
    model.fit(X)

    # shapiro_alpha=0.1
    model = RCD(shapiro_alpha=0.1)
    model.fit(X)

    # shapiro_alpha=0.1
    model = RCD(shapiro_alpha=0.0)
    model.fit(X)

    # MLHSICR=True
    model = RCD(MLHSICR=True)
    model.fit(X)

    # bw_method='scott'
    model = RCD(bw_method='scott')
    model.fit(X)

    # bw_method='silverman'
    model = RCD(bw_method='silverman')
    model.fit(X)

    # no latent confounders
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x0 = get_external_effect(n_samples)
    x3 = get_external_effect(n_samples)
    x1 = 1.0*x0 + 1.0*x3 + get_external_effect(n_samples)
    x2 = 0.8*x1 + get_external_effect(n_samples)
    x4 = 1.0*x1 + get_external_effect(n_samples)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])
    model = RCD()
    model.fit(X)
    p_values = model.get_error_independence_p_values(X)

    # causal direction: x3-->x0, x3-->x1, x0,x1-->x2
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x3 = get_external_effect(n_samples)
    x0 = 0.5*x3 + get_external_effect(n_samples)
    x1 = 0.5*x3 + get_external_effect(n_samples)
    x2 = 1.0*x0 + 1.0*x1 + get_external_effect(n_samples)
    X = pd.DataFrame(np.array([x0, x1, x2]).T, columns=['x0', 'x1', 'x2'])
    model = RCD()
    model.fit(X)
    p_values = model.get_error_independence_p_values(X)



def test_fit_invalid_data():
    # Not array data
    X = 1
    try:
        model = RCD()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = RCD()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=100)
    x1 = 2.0*x0 + np.random.uniform(size=100)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[10, 0] = np.nan
    try:
        model = RCD()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=100)
    x1 = 2.0*x0 + np.random.uniform(size=100)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[10, 0] = np.inf
    try:
        model = RCD()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: max_explanatory_num
    x0 = np.random.uniform(size=100)
    x1 = 2.0*x0 + np.random.uniform(size=100)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = RCD(max_explanatory_num=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: cor_alpha
    try:
        model = RCD(cor_alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: ind_alpha
    try:
        model = RCD(ind_alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: shapiro_alpha
    try:
        model = RCD(shapiro_alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: bw_method
    try:
        model = RCD(bw_method='X')
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x5 = get_external_effect(n_samples)
    x6 = get_external_effect(n_samples)
    x0 = 0.6*x5 + get_external_effect(n_samples)
    x3 = 0.5*x5 + get_external_effect(n_samples)
    x1 = 1.0*x0 + 1.0*x3 + get_external_effect(n_samples)
    x2 = 0.8*x1 - 0.6*x6 + get_external_effect(n_samples)
    x4 = 1.0*x1 - 0.5*x6 + get_external_effect(n_samples)
    # x5 and x6 are latent confounders
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T,
                     columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    model = RCD()
    result = model.bootstrap(X, n_sampling=20)

    result.adjacency_matrices_
    result.total_effects_

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # MLHSICR=True
    model = RCD(MLHSICR=True)
    result = model.bootstrap(X, n_sampling=20)

    # no latent confounders
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x0 = get_external_effect(n_samples)
    x3 = get_external_effect(n_samples)
    x1 = 1.0*x0 + 1.0*x3 + get_external_effect(n_samples)
    x2 = 0.8*x1 + get_external_effect(n_samples)
    x4 = 1.0*x1 + get_external_effect(n_samples)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    model = RCD()
    result = model.bootstrap(X, n_sampling=20)

    # causal direction: x3-->x0, x3-->x1, x0,x1-->x2
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x3 = get_external_effect(n_samples)
    x0 = 0.5*x3 + get_external_effect(n_samples)
    x1 = 0.5*x3 + get_external_effect(n_samples)
    x2 = 1.0*x0 + 1.0*x1 + get_external_effect(n_samples)
    X = pd.DataFrame(np.array([x0, x1, x2]).T, columns=['x0', 'x1', 'x2'])

    model = RCD()
    result = model.bootstrap(X, n_sampling=20)

def test_bootstrap_invalid():
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x5 = get_external_effect(n_samples)
    x6 = get_external_effect(n_samples)
    x0 = 0.6*x5 + get_external_effect(n_samples)
    x3 = 0.5*x5 + get_external_effect(n_samples)
    x1 = 1.0*x0 + 1.0*x3 + get_external_effect(n_samples)
    x2 = 0.8*x1 - 0.6*x6 + get_external_effect(n_samples)
    x4 = 1.0*x1 - 0.5*x6 + get_external_effect(n_samples)
    # x5 and x6 are latent confounders
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T,
                     columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    model = RCD()

    # Invalid value: n_sampling=-1
    try:
        result = model.bootstrap(X, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: n_sampling='3'
    try:
        result = model.bootstrap(X, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError



def test_estimate_total_effect_invalid():
    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3
    n_samples = 100
    x5 = get_external_effect(n_samples)
    x6 = get_external_effect(n_samples)
    x0 = 0.6*x5 + get_external_effect(n_samples)
    x3 = 0.5*x5 + get_external_effect(n_samples)
    x1 = 1.0*x0 + 1.0*x3 + get_external_effect(n_samples)
    x2 = 0.8*x1 - 0.6*x6 + get_external_effect(n_samples)
    x4 = 1.0*x1 - 0.5*x6 + get_external_effect(n_samples)
    # x5 and x6 are latent confounders
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T,
                     columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    model = RCD()
    model.fit(X)
    model._ancestors_list[1] = [2]
    model._ancestors_list[2] = [3, 4]
    model._adjacency_matrix = [[0, np.nan], [0, np.nan], [0, np.nan]]

    # warning
    model.estimate_total_effect(X, 1, 2)
    model.estimate_total_effect(X, 2, 1)



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""

import numpy as np
import scipy.optimize as sopt
from scipy.special import expit as sigmoid
from factor_analyzer import (ConfirmatoryFactorAnalyzer, ModelSpecificationParser)
from statsmodels.stats.outliers_influence import variance_inflation_factor
import pandas as pd


class LiNA:
    """Implementation of LiNA Algorithm [1]_

    References
    ----------
    .. [1] Y. Zeng, S. Shimizu, R. Cai, F. Xie, M. Yamamoto and Z. Hao.
       Causal Discovery with Multi-Domain LiNGAM for Latent Factors.
       In Proc. of the Thirtieth International Joint Conference on
       Artificial Intelligence (IJCAI-21), 2021: 2097--2103.
    """

    def __init__(
            self,
            w_threshold=0.3,
            lambda1=0.1,
            lambda2=0.1,
            loss_type='laplace',
            max_iter=100,
            h_tol=1e-8,
            rho_max=1e+16
    ):
        """Construct a LiNA model.

        Parameters
        ----------
         w_threshold : float (default=0.3)
             Drop edge if the weight btw. latent factors is less than w_threshold.
         lambda1 : float, optional (default=0.1)
             L1 penalty parameter.
         lambda2 : float, (default=0.1)
             L2 penalty parameter.
         loss_type : str, (default='laplace')
             Type of distribution of the noise.
         max_iter : int, (default=100)
             Maximum number of dual ascent steps.
         h_tol : float, (default=1e-8)
             Tolerance parameter of the acyclicity constraint.
         rho_max : float, (default=1e+16)
             Maximum value of the regularization parameter rho.
        """

        self._w_threshold = w_threshold
        self._lambda1 = lambda1
        self._lambda2 = lambda2
        self._loss_type = loss_type
        self._max_iter = max_iter
        self._h_tol = h_tol
        self._rho_max = rho_max
        self._adjacency_matrix = None
        self._measurement_matrix = None

    def fit(self, X, G_sign, scale):
        """Fit the model to X with measurement structure and latent factors' scales.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of measurement features.
        G_sign: array-like, shape (n_features, n_features_latent)
            Measurement structure matrix, where ``n_features_latent`` is
            the number of  latent factors and ``n_features`` is the
            number of measurement features.
        scale: array-like, shape (1, n_features_latent)
            Scales of the latent factors.

        Returns
        -------
        self : object
            Returns the instance of self.
        """
        # X = check_array(X)
        n_samples = X.shape[0]
        n_features_latent = G_sign.shape[1]

        # structure of G is known, noises e are ignored
        model_dict = {"F1": ["V1", "V2"],
                      "F2": ["V3", "V4"],
                      "F3": ["V5", "V6"],
                      "F4": ["V7", "V8"],
                      "F5": ["V9", "V10"]}
        # "F6": ["V11", "V12"],
        # "F7": ["V13", "V14"],
        # "F8": ["V15", "V16"],
        # "F9": ["V17", "V18"],
        # "F10": ["V19", "V20"]}

        model_spec = ModelSpecificationParser.parse_model_specification_from_dict(X, model_dict)

        cfa = ConfirmatoryFactorAnalyzer(model_spec, disp=True, max_iter=100000)
        cfa.fit(X, G_sign)
        G_est_nor = cfa.loadings_
        # print(G_est_nor)

        f_est_nor = np.linalg.inv(G_est_nor.T @ G_est_nor) @ G_est_nor.T @ (X.T)
        f_est_nor = f_est_nor.T
        f_est = np.zeros([n_samples, n_features_latent])
        for j in range(n_features_latent):
            f_est[:, j] = f_est_nor[:, j] * scale[0, j]

        # W_est = _estimate_LiNA(f_est, self._w_threshold, self._lambda1, self._loss_type)  # input f: n*d
        f = f_est
        W_est = self._estimate_LiNA(f)
        self._adjacency_matrix = W_est
        self._measurement_matrix = G_est_nor
        return self._adjacency_matrix

    def _estimate_LiNA(self, f):
        """Estimate the adjacency matrix btw. estimated latent factors for single-domain data"""

        def _loss(W):
            """Evaluate value and gradient of loss."""
            M = f @ W
            if self._loss_type == 'l2':
                R = f - M
                loss = 0.5 / f.shape[0] * (R ** 2).sum()
                D = - 1.0 / f.shape[0] * f.T @ R
            elif self._loss_type == 'logistic':
                loss = 1.0 / f.shape[0] * (np.logaddexp(0, M) - f * M).sum()
                D = 1.0 / f.shape[0] * f.T @ (sigmoid(M) - f)
            elif self._loss_type == 'poisson':
                S = np.exp(M)
                loss = 1.0 / f.shape[0] * (S - f * M).sum()
                D = 1.0 / f.shape[0] * f.T @ (S - f)
            elif self._loss_type == 'laplace':
                R = f - M
                loss = - np.sum(-np.log(np.cosh(R)))
                D = - f.T @ np.tanh(R)
            else:
                raise ValueError('unknown loss type')
            return loss, D

        def _h(W):
            """Evaluate value and gradient of acyclicity constraint."""
            #     E = slin.expm(W * W)  # (Zheng et al. 2018)
            #     h = np.trace(E) - d
            M = np.eye(d) + W * W / d  # (Yu et al. 2019)
            E = np.linalg.matrix_power(M, d - 1)
            h = (E.T * M).sum() - d
            return h, E

        def _adj(w):
            """Convert doubled variables ([2 d^2] array) back to original variables ([d, d] matrix)."""
            return (w[:d * d] - w[d * d:]).reshape([d, d])

        def _func(w):
            """Evaluate value and gradient of augmented Lagrangian for doubled variables ([2 d^2] array)."""
            W = _adj(w)
            loss, D = _loss(W)
            h, E = _h(W)
            obj = loss + 0.5 * rho * h * h + alpha * h + self._lambda1 * w.sum()
            # obj = loss + 0.5 * rho * h * h + alpha * h + self._lambda1 * w.sum()
            # + self._lambda2*(np.linalg.norm(W, ord = 'fro')**2) # add l2 penalty constraint
            G = D + (rho * h + alpha) * E.T * W * 2
            # G = D + (rho * h + alpha) * E.T * W * 2 + 2 * self._lambda2 * W #gradient of loss + h + penalty
            grad_cat = np.concatenate((G + self._lambda1, - G + self._lambda1), axis=None)
            return obj, grad_cat

        n, d = f.shape
        w_est, rho, alpha, h = np.zeros(2 * d * d), 1.0, 0.0, np.inf  # double w_est into (w_pos, w_neg)
        bnds = [(0, 0) if i == j else (0, None) for _ in range(2) for i in range(d) for j in range(d)]
        for _ in range(self._max_iter):
            w_new, h_new = None, None
            while rho < self._rho_max:
                sol = sopt.minimize(_func, w_est, method='L-BFGS-B', jac=True, bounds=bnds)
                w_new = sol.x
                h_new, _ = _h(_adj(w_new))
                if h_new > 0.25 * h:
                    rho *= 10
                else:
                    break
            w_est, h = w_new, h_new
            alpha += rho * h
            if h <= self._h_tol or rho >= self._rho_max:
                break
        W_est = _adj(w_est)
        W_est[np.abs(W_est) < self._w_threshold] = 0

        return W_est

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix between latent factors.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features_latent, n_features_latent)
            The adjacency matrix of latent factors, where
            ``n_features_latent`` is the number of latent
            factors.
        """
        return self._adjacency_matrix

    @property
    def measurement_matrix_(self):
        """Estimated measurement matrix between measurement variables and
        latent factors.

        Returns
        -------
        measurement_matrix_ : array-like, shape (n_features, n_features_latent)
            The measurement matrix between measurement variables and
            latent factors, where ``n_features_latent`` is the
            number of latent factors and ``n_features`` is the
            number of measurement variables.
        """
        return self._measurement_matrix


class MDLiNA:
    """Implementation of MD-LiNA Algorithm [1]_

    References
    ----------
    .. [1] Y. Zeng, S. Shimizu, R. Cai, F. Xie, M. Yamamoto and Z. Hao.
       Causal Discovery with Multi-Domain LiNGAM for Latent Factors.
       In Proc. of the Thirtieth International Joint Conference on
       Artificial Intelligence (IJCAI-21), 2021: 2097--2103.
    """

    def __init__(
            self,
            w_threshold=0.3,
            lambda1=0.1,
            lambda2=0.1,
            loss_type='laplace',
            max_iter=100,
            h_tol=1e-8,
            rho_max=1e+16,
            no_of_domain=2,
            no_of_latent_1domain=3,
    ):
        """Construct an MD-LiNA model.

        Parameters
        ----------
         w_threshold : float (default=0.3)
             Drop edge if the weight btw. latent factors is less than w_threshold.
         lambda1 : float, optional (default=0.1)
             L1 penalty parameter.
         lambda2 : float, (default=0.1)
             L2 penalty parameter.
         loss_type : str, (default='laplace')
             Type of distribution of the noise.
         max_iter : int, (default=100)
             Maximum number of dual ascent steps.
         h_tol : float, (default=1e-8)
             Tolerance parameter of the acyclicity constraint.
         rho_max : float, (default=1e+16)
             Maximum value of the regularization parameter rho.
         no_of_domain : int, (default=2)
             Number of domains.
         no_of_latent_1domain : float, (default=3)
             Number of latent factors in a domain.

        """

        self._w_threshold = w_threshold
        self._lambda1 = lambda1
        self._lambda2 = lambda2
        self._loss_type = loss_type
        self._max_iter = max_iter
        self._h_tol = h_tol
        self._rho_max = rho_max
        self._no_of_domain = no_of_domain
        self._no_of_latent_1domain = no_of_latent_1domain
        self._adjacency_matrix = None
        self._measurement_matrix = None

    def fit(self, X, G_sign, scale):
        """Fit the model to X with measurement structure and latent factors' scales.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            of all domains and ``n_features`` is the number of features
            of all domains.
        G_sign: array-like, shape (n_features, n_features_latent)
            Measurement structure matrix, where ``n_features_latent`` is
            the number of latent factors of all domains and ``n_features``
            is the number of measurement variables of all domains.
        scale: array-like, shape (1, n_features_latent)
            Scales of the latent factors.

        Returns
        -------
        self : object
            Returns the instance of self.
        """

        # PHASE I: use CFA to estimate \bar{G}
        model_dict = {"F1": ["V1", "V2"],
                      "F2": ["V3", "V4"],
                      "F3": ["V5", "V6"],
                      "F4": ["V7", "V8"],
                      "F5": ["V9", "V10"],
                      "F6": ["V11", "V12"]}

        model_spec = ModelSpecificationParser.parse_model_specification_from_dict(X, model_dict)
        cfa = ConfirmatoryFactorAnalyzer(model_spec, disp=True, max_iter=100000)
        # for better estimation, we input G_sign as well to help initialize the loading matrix
        cfa.fit(X, G_sign)
        G_est = cfa.loadings_

        f_est = np.linalg.inv(G_est.T @ G_est) @ G_est.T @ (X.T)
        f_est = f_est.T
        # F = f_est

        F = np.zeros([f_est.shape[0], f_est.shape[1]])
        for j in range(f_est.shape[1]):
            if j < self._no_of_latent_1domain:
                F[:, j] = f_est[:, j] * scale[0, j]
            else:
                F[:, j] = f_est[:, j] * scale[1, j]

        # optimize together
        W0 = np.zeros([self._no_of_latent_1domain, self._no_of_latent_1domain])
        H0 = np.zeros([self._no_of_latent_1domain * self._no_of_domain, self._no_of_latent_1domain])
        times = 0
        min_like = np.inf
        Likelihood_all = np.zeros([1, 10000])

        while 1:
            # update H
            H_est = self._estimate_MDLiNA_H(F, W0)
            F_final2 = np.linalg.inv(H_est.T @ H_est) @ H_est.T @ F.T

            # update B
            W_est = self._estimate_MDLiNA_B(F_final2.T)
            times = times + 1

            print('W_est is:\n', W_est)
            print('W0 is:\n', W0)

            # check if convergent?
            inv_HH = np.linalg.inv(H_est.T @ H_est) @ H_est.T
            P_H = H_est @ inv_HH @ F.T
            LL = np.sum(np.log(np.cosh(F @ H_est - F @ inv_HH.T @ W_est))) + (np.abs(W_est)).sum() + 10000 * self._conh(
                W_est)
            recon_error = np.trace((F.T - P_H).T @ (F.T - P_H))
            H1 = (np.abs(H_est)).sum()
            likelihood_all = LL + recon_error + H1
            Likelihood_all[0, times] = likelihood_all
            if likelihood_all <= min_like:
                min_like = likelihood_all
            if np.abs(Likelihood_all[0, times] - Likelihood_all[0, times - 1]) < 1e-1:
                break
            elif times >= 30 and np.abs(likelihood_all - min_like) < 1e-1:
                break
            elif times >= 30 and likelihood_all <= min_like:
                break
            elif times >= 50 and np.abs(likelihood_all - np.min(Likelihood_all[0, times - 20:times - 1])) < 1e-1:
                break
            elif times >= 50 and likelihood_all <= np.min(Likelihood_all[0, times - 20:times - 1]):
                break
            else:
                W0 = W_est
                H0 = H_est

        if np.where(H_est != 0)[0].shape[0] == 0:
            H_est = H0
        if np.where(W_est != 0)[0].shape[0] == 0:
            W_est = W0

        # choose new factors: represent which concept.
        tt = 0
        if tt == 0:
            Weight = np.abs(H_est)
            W = W_est
        else:
            Weight = np.abs(H_est)
            W = W0
        BB = np.zeros([self._no_of_latent_1domain, self._no_of_latent_1domain])
        for i in range(self._no_of_latent_1domain):
            for j in range(self._no_of_latent_1domain):
                BB[j, i] = np.abs(Weight[j, i]) + np.abs(Weight[j + self._no_of_latent_1domain, i])
        order = np.zeros([1, self._no_of_latent_1domain])
        if len(np.where(BB == 0)) == 0:
            for i in range(self._no_of_latent_1domain):
                indice = np.where(BB == np.amax(BB))
                order[0, indice[1]] = indice[0]
                BB[indice[0], :] = 0
                BB[:, indice[1]] = 0
        else:
            indice = np.where(BB == 0)
            BB[:, indice[1]] = BB[:, indice[1]] + 0.01
            for i in range(self._no_of_latent_1domain):
                indice = np.where(BB == np.amax(BB))
                if len(indice[0]) > 1:
                    order[0, indice[1][0]] = indice[0][0]
                    BB[indice[0][0], :] = 0
                    BB[:, indice[1][0]] = 0
                else:
                    order[0, indice[1]] = indice[0]
                    BB[indice[0], :] = 0
                    BB[:, indice[1]] = 0

        W_est_new0 = np.zeros([self._no_of_latent_1domain, self._no_of_latent_1domain])
        W_est_new = np.zeros([self._no_of_latent_1domain, self._no_of_latent_1domain])
        for i in range(self._no_of_latent_1domain):
            indice = np.where(order == i)
            W_est_new0[[i], :] = W[indice[1], :]  # row
        for i in range(self._no_of_latent_1domain):
            indice = np.where(order == i)
            W_est_new[:, [i]] = W_est_new0[:, indice[1]]  # column

        self._adjacency_matrix = W_est_new
        self._measurement_matrix = G_est

        return self._adjacency_matrix

    def calc_vif(X):
        # Calculating VIF
        vif = pd.DataFrame()
        X = pd.DataFrame(X)  # convert ndarray to dataframe
        vif["variables"] = X.columns
        vif["VIF"] = [variance_inflation_factor(X.values, i) for i in range(X.shape[1])]

        return (vif)

    def _estimate_MDLiNA_H(self, f, W):
        """ Optimize H matrix btw. multi-domain latent factors and latent factors of interest

            Returns:
                H_est (np.ndarray): [hq, q] estimated H
            """

        def _loss(H):
            """Evaluate value and gradient of loss."""
            inv_HH = np.linalg.inv(H.T @ H)
            fH = f @ H
            M = fH @ inv_HH @ W
            if self._loss_type == 'l2':
                R = f - M
                loss = 0.5 / f.shape[0] * (R ** 2).sum()
                D = - 1.0 / f.shape[0] * f.T @ R
            elif self._loss_type == 'laplace':
                R = fH @ inv_HH - M
                # standardize
                for j in range(hq):
                    R[:, j] = R[:, j] - np.mean(R[:, j])
                    R[:, j] = R[:, j] / np.std(R[:, j])
                if np.trace(H.T @ H) < 10000:
                    aa = np.trace(H.T @ H)
                    D = - f.T @ np.tanh(R) @ (np.eye(np.size(H, 1)) - W) @ inv_HH \
                        + H @ inv_HH @ inv_HH @ fH.T @ np.tanh(R) @ (np.eye(np.size(H, 1)) - W) \
                        + H @ inv_HH @ inv_HH @ (np.eye(np.size(H, 1)) - W).T @ np.tanh(R).T @ fH \
                        - 2 * f.T @ fH @ inv_HH + 2 * H @ inv_HH @ fH.T @ fH @ inv_HH
                else:
                    aa = 0
                    D = - f.T @ np.tanh(R) @ (np.eye(np.size(H, 1)) - W) @ inv_HH \
                        + H @ inv_HH @ inv_HH @ fH.T @ np.tanh(R) @ (np.eye(np.size(H, 1)) - W) \
                        + H @ inv_HH @ inv_HH @ (np.eye(np.size(H, 1)) - W).T @ np.tanh(R).T @ fH \
                        - 2 * f.T @ fH @ inv_HH + 2 * H @ inv_HH @ fH.T @ fH @ inv_HH + 2 * H
                loss = np.sum(np.log(np.cosh(R))) \
                    + np.trace((f - fH @ inv_HH @ H.T).T @ (f - fH @ inv_HH @ H.T)) \
                    + (np.trace(H.T @ H) - aa)
            else:
                raise ValueError('unknown loss type')
            return loss, D

        def _adj(h):
            """Convert doubled variables ([2 d^2] array) back to original variables ([d, d] matrix)."""
            return (h[:hq * q] - h[hq * q:]).reshape([q, hq])

        def _func(h):
            """Evaluate value and gradient of augmented Lagrangian for doubled variables ([2 d^2] array)."""
            H = _adj(h)
            loss, D = _loss(H)
            obj = loss + self._lambda1 * h.sum()
            G = D  # gradient of loss
            grad_cat = np.concatenate((G + self._lambda1, - G + self._lambda1), axis=None)
            return obj, grad_cat

        n, q = f.shape
        hq = np.size(W, 0)  # hq is \hat{q}: the number of shared latent factors
        # ut.set_random_seed(0)
        h_est, score_0 = np.random.rand(2 * hq * q), np.inf
        for _ in range(self._max_iter):
            # h_new, score_1 = None, None
            sol = sopt.minimize(_func, h_est, method='BFGS', jac=True)
            h_new = sol.x
            score_1 = sol.fun

            if np.abs(score_0 - score_1) < self._h_tol:
                break
            if score_1 > 0 and score_1 < score_0:
                h_est = h_new
                score_0 = score_1
                # see if singular?
                bb = _adj(h_est)
                bb[np.abs(bb) < self._w_threshold] = 0
                if np.linalg.det(bb.T @ bb) == 0:
                    h_est[0:8] = h_est[0:8] + 0.1
                if np.max(np.abs(h_est)) > 100:
                    h_est = np.random.rand(2 * hq * q)
                    score_0 = np.inf
            else:
                h_est = np.random.rand(2 * hq * q)

        h_est = h_new
        H_est = _adj(h_est)
        H_est[np.abs(H_est) < self._w_threshold] = 0
        self._H_est = H_est
        return self._H_est

    def _estimate_MDLiNA_B(self, f):
        """Optimize the causal effect matrix among shared latent factors

        Returns:
            W_est (np.ndarray): [d, d]
        """

        def _loss(W):
            """Evaluate value and gradient of loss."""
            M = f @ W
            if self._loss_type == 'l2':
                R = f - M
                loss = 0.5 / f.shape[0] * (R ** 2).sum()
                D = - 1.0 / f.shape[0] * f.T @ R
            elif self._loss_type == 'logistic':
                loss = 1.0 / f.shape[0] * (np.logaddexp(0, M) - f * M).sum()
                D = 1.0 / f.shape[0] * f.T @ (sigmoid(M) - f)
            elif self._loss_type == 'poisson':
                S = np.exp(M)
                loss = 1.0 / f.shape[0] * (S - f * M).sum()
                D = 1.0 / f.shape[0] * f.T @ (S - f)
            elif self._loss_type == 'laplace':
                R = f - M
                loss = - np.sum(-np.log(np.cosh(R)))
                D = - f.T @ np.tanh(R)
            else:
                raise ValueError('unknown loss type')
            return loss, D

        def _h(W):
            """Evaluate value and gradient of acyclicity constraint."""
            M = np.eye(d) + W * W / d  # (Yu et al. 2019)
            E = np.linalg.matrix_power(M, d - 1)
            h = (E.T * M).sum() - d
            return h, E

        def _adj(w):
            """Convert doubled variables ([2 d^2] array) back to original variables ([d, d] matrix)."""
            return (w[:d * d] - w[d * d:]).reshape([d, d])

        def _func(w):
            """Evaluate value and gradient of augmented Lagrangian for doubled variables ([2 d^2] array)."""
            W = _adj(w)
            loss, D = _loss(W)
            h, E = _h(W)
            obj = loss + 0.5 * rho * h * h + alpha * h + self._lambda1 * w.sum()
            G = D + (rho * h + alpha) * E.T * W * 2
            grad_cat = np.concatenate((G + self._lambda1, - G + self._lambda1), axis=None)
            return obj, grad_cat

        n, d = f.shape
        w_est, rho, alpha, h = np.zeros(2 * d * d), 1.0, 0.0, np.inf
        bnds = [(0, 0) if i == j else (0, None) for _ in range(2) for i in range(d) for j in range(d)]
        for _ in range(self._max_iter):
            w_new, h_new = None, None
            while rho < self._rho_max:
                sol = sopt.minimize(_func, w_est, method='L-BFGS-B', jac=True, bounds=bnds)
                w_new = sol.x
                h_new, _ = _h(_adj(w_new))
                if h_new > 0.25 * h:
                    rho *= 10
                else:
                    break
            w_est, h = w_new, h_new
            alpha += rho * h
            if h <= self._h_tol or rho >= self._rho_max:
                break
        W_est = _adj(w_est)
        W_est[np.abs(W_est) < self._w_threshold] = 0
        self._W_est = W_est
        return self._W_est

    def _conh(self, W):
        """Evaluate value of acyclicity constraint."""
        d = self._no_of_latent_1domain
        M = np.eye(d) + W * W / d  # (Yu et al. 2019)
        E = np.linalg.matrix_power(M, d - 1)
        h = (E.T * M).sum() - d
        return h

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix between latent factors of interest, which
        is shared by all domains.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features_latent_1domain, n_features_latent_1domain)
            The adjacency matrix of latent factors of interest, where
            ``n_features_latent_1domain`` is the number of latent
            factors of interest.
        """
        return self._adjacency_matrix

    @property
    def measurement_matrix_(self):
        """Estimated measurement matrix between measurement variables and
        latent factors from all domains.

        Returns
        -------
        measurement_matrix_ : array-like, shape (n_features, n_features_latent)
            The measurement matrix between measurement variables and
            latent factors, where ``n_features_latent`` is the
            number of latent factors and ``n_features`` is the
            number of measurement variables from all domains.
        """
        return self._measurement_matrix



import numpy as np
import pandas as pd

from lingam.utils import (
    print_causal_directions,
    print_dagc,
    make_prior_knowledge,
    remove_effect,
    make_dot,
    predict_adaptive_lasso,
    get_sink_variables,
    get_exo_variables,
    find_all_paths,
)


def test_print_causal_directions():
    cdc = {'from': [0, 1, 2, 3], 'to': [1, 2, 3, 4], 'count': [5, 4, 3, 2]}
    print_causal_directions(cdc, n_sampling=5)
    print_causal_directions(cdc, n_sampling=5, labels=['A', 'B', 'C', 'D', 'E'])


def test_print_dagc():
    dagc = {
        'dag': [
            {'from': [0, 1, 2, 3], 'to': [1, 2, 3, 4]},
            {'from': [0, 1, 2, 3], 'to': [1, 2, 3, 0]},
            {'from': [0, 1, 2, 3], 'to': [1, 2, 3, 4]}
        ],
        'count': [5, 2, 1],
    }
    print_dagc(dagc, n_sampling=5)
    print_dagc(dagc, n_sampling=5, labels=['A', 'B', 'C', 'D', 'E'])

def test_make_prior_knowledge():
    pk = make_prior_knowledge(n_variables=5, exogenous_variables=[0], sink_variables=[1], paths=[[2, 3]], no_paths=[[3, 4]])
    assert pk.shape[0]==5 and pk.shape[1]==5

def test_get_sink_variables():
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 1, 0, 0],
    ]
    sv = get_sink_variables(adjacency_matrix=am)
    assert len(sv)==2

def test_get_exo_variables():
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    ev = get_exo_variables(adjacency_matrix=am)
    assert len(ev)==2

def test_remove_effect():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    remove_effect(X, remove_features=[0])

def test_make_dot():
    # default
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    ev = make_dot(am, labels=['x0', 'x1', 'x2', 'x3'])

    # default
    am = [
        [0, 1, 1, 0],
        [0, 0, np.nan, 0],
        [0, np.nan, 0, 0],
        [0, 1, 0, 0],
    ]
    ev = make_dot(am, labels=['x0', 'x1', 'x2', 'x3'])

    # prediction
    ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_coefs=[0.1, 0.1])
    ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_target_label='Target', prediction_line_color='#0000FF')
    ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_feature_importance=[0.5, 0.2])

    # invalid adjacency matrix
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
    ]
    try:
        ev = make_dot(am, labels=['x0', 'x1', 'x2', 'x3'])
    except ValueError:
        pass
    else:
        raise AssertionError

    # invalid label size
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    try:
        ev = make_dot(am, labels=['x0', 'x1', 'x2'])
    except ValueError:
        pass
    else:
        raise AssertionError

    # invalid predict settings
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    try:
        ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_coefs=[0.1])
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_feature_importance=[0.5])
    except ValueError:
        pass
    else:
        raise AssertionError



def test_predict_adaptive_lasso():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    predict_adaptive_lasso(X.values, predictors=[0, 1, 2], target=3)

def test_find_all_paths():
    dag = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 1, 0, 0],
    ]
    find_all_paths(np.array(dag), 1, 0)

    # min_causal_effect
    dag = [
        [0, 2, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 1, 0, 0],
    ]
    find_all_paths(np.array(dag), 1, 0, min_causal_effect=1)

    # Invalid DAG: cycle
    dag = [
        [1, 1, 1, 1],
        [1, 1, 1, 1],
        [1, 1, 1, 1],
        [1, 1, 1, 1],
    ]
    try:
        find_all_paths(np.array(dag), 1, 0)
    except ValueError:
        pass
    else:
        raise AssertionError



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""

import itertools
import numbers
import warnings

import numpy as np
from scipy.stats.distributions import chi2
from sklearn.utils import check_array, resample

from .bootstrap import BootstrapResult
from .hsic import hsic_test_gamma
from .utils import predict_adaptive_lasso


class BottomUpParceLiNGAM:
    """Implementation of ParceLiNGAM Algorithm [1]_

    References
    ----------
    .. [1] T. Tashiro, S. Shimizu, and A. Hyvärinen.
       ParceLiNGAM: a causal ordering method robust against latent confounders.
       Neural computation, 26.1: 57-83, 2014.
    """

    def __init__(
        self, random_state=None, alpha=0.1, regressor=None, prior_knowledge=None
    ):
        """Construct a BottomUpParceLiNGAM model.

        Parameters
        ----------
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        alpha : float, optional (default=0.1)
            Significant level of statistical test. If alpha=0.0, rejection does not occur in statistical tests.
        regressor : regressor object implementing 'fit' and 'predict' function (default=None)
            Regressor to compute residuals.
            This regressor object must have ``fit``method and ``predict`` function like scikit-learn's model.
        prior_knowledge : array-like, shape (n_features, n_features), optional (default=None)
            Prior knowledge used for causal discovery, where ``n_features`` is the number of features.

            The elements of prior knowledge matrix are defined as follows [1]_:

            * ``0`` : :math:`x_i` does not have a directed path to :math:`x_j`
            * ``1`` : :math:`x_i` has a directed path to :math:`x_j`
            * ``-1`` : No prior knowledge is available to know if either of the two cases above (0 or 1) is true.
        """
        # Check parameters
        if regressor is not None:
            if not (hasattr(regressor, "fit") and hasattr(regressor, "predict")):
                raise ValueError("'regressor' has no fit or predict method.")

        if alpha < 0.0:
            raise ValueError("alpha must be an float greater than 0.")

        self._random_state = random_state
        self._alpha = alpha
        self._causal_order = None
        self._adjacency_matrix = None
        self._reg = regressor
        self._Aknw = prior_knowledge

        if self._Aknw is not None:
            self._Aknw = check_array(self._Aknw)
            self._Aknw = np.where(self._Aknw < 0, np.nan, self._Aknw)

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._causal_order = None
        self._adjacency_matrices = None

        # Check parameters
        X = check_array(X)
        n_features = X.shape[1]

        # Check prior knowledge
        if self._Aknw is not None:
            if (n_features, n_features) != self._Aknw.shape:
                raise ValueError(
                    "The shape of prior knowledge must be (n_features, n_features)"
                )
            else:
                # Extract all partial orders in prior knowledge matrix
                self._partial_orders = self._extract_partial_orders(self._Aknw)

        # Center variables for each group
        X = X - np.tile(np.mean(X, axis=0), (X.shape[0], 1))

        # bonferroni correction
        thresh_p = self._alpha / (n_features - 1)

        # Search causal orders one by one from the bottom upward
        K_bttm, p_bttm = self._search_causal_order(X, thresh_p)

        U_res = list(np.setdiff1d(np.arange(n_features), K_bttm))
        K = []
        # Add a list of features whose order is unknown.
        if len(U_res) > 1:
            K = [U_res]
        for i in K_bttm:
            K.append(i)

        self._causal_order = K
        self._p_list = p_bttm
        return self._estimate_adjacency_matrix(X, prior_knowledge=self._Aknw)

    def _extract_partial_orders(self, pk):
        """Extract partial orders from prior knowledge."""
        path_pairs = np.array(np.where(pk == 1)).transpose()
        no_path_pairs = np.array(np.where(pk == 0)).transpose()

        # Check for inconsistencies in pairs with path
        check_pairs = np.concatenate([path_pairs, path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            if len(pairs[counts > 1]) > 0:
                raise ValueError(
                    f"The prior knowledge contains inconsistencies at the following indices: {pairs[counts>1].tolist()}"
                )

        # Check for inconsistencies in pairs without path
        # If there are duplicate pairs without path, they cancel out and are not ordered.
        check_pairs = np.concatenate([no_path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            check_pairs = np.concatenate([no_path_pairs, pairs[counts > 1]])
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            no_path_pairs = pairs[counts < 2]

        check_pairs = np.concatenate([path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) == 0:
            # If no pairs are extracted from the specified prior knowledge,
            return check_pairs

        pairs = np.unique(check_pairs, axis=0)
        return pairs[:, [1, 0]]  # [to, from] -> [from, to]

    def _search_candidate(self, U):
        """Search for candidate features"""
        # If no prior knowledge is specified, nothing to do.
        if self._Aknw is None:
            return U

        # Candidate features that are not to the left of the partial orders
        if len(self._partial_orders) != 0:
            Uc = [i for i in U if i not in self._partial_orders[:, 0]]
            return Uc

        return U

    def _search_causal_order(self, X, thresh_p):
        """Search causal orders one by one from the bottom upward."""
        U = np.arange(X.shape[1])
        K_bttm = []
        p_bttm = []
        is_search_causal_order = True

        while is_search_causal_order:
            # Search for candidate features
            Uc = self._search_candidate(U)

            if len(Uc) == 1:
                # If there is only one variable in Uc,
                # calculate HSIC with the rest of the variables
                m = np.array([Uc[0]])
                predictors = np.setdiff1d(U, Uc[0])
                R = self._compute_residuals(X, predictors, m)
                fisher_p, _ = self._fisher_hsic_test(X[:, predictors], R, np.inf)
            else:
                # Find the most sink variable
                m, _, fisher_p = self._find_exo_vec(X, Uc)

            # Conduct statistical test by the p-value or the statistic
            # If statistical test is not rejected
            if fisher_p >= thresh_p:
                # Add index of the exogenous variable to K_bttm
                K_bttm = np.append(m, K_bttm).astype(np.int64)
                p_bttm.insert(0, fisher_p)

                # Update U and partial orders
                U = U[U != m]
                if self._Aknw is not None:
                    self._partial_orders = self._partial_orders[
                        self._partial_orders[:, 1] != m
                    ]

                # If there is only one candidate for sink variable, the search ends
                if len(U) <= 1:
                    K_bttm = np.append(U, K_bttm).astype(np.int64)
                    p_bttm.insert(0, 0.0)
                    is_search_causal_order = False
            # If statistical test is rejected
            else:
                is_search_causal_order = False

        return K_bttm, p_bttm

    def _find_exo_vec(self, X, U):
        """Find the most exogenous vector."""
        max_p = -np.inf
        max_p_stat = np.inf

        exo_vec = []
        for j in range(len(U)):
            xi_index = np.setdiff1d(U, U[j])
            xj_index = np.array([U[j]])

            # Compute residuals
            R = self._compute_residuals(X, xi_index, xj_index)

            # HSIC test with Fisher's method
            fisher_p, fisher_stat = self._fisher_hsic_test(
                X[:, xi_index], R, max_p_stat
            )

            # Update output
            if fisher_stat < max_p_stat or fisher_p > max_p:
                exo_vec = xi_index
                max_p = fisher_p
                max_p_stat = fisher_stat

        m = np.setdiff1d(U, exo_vec)
        return m, exo_vec, max_p

    def _compute_residuals(self, X, predictors, target):
        """Compute residuals"""
        if self._reg is None:
            # Compute residuals of least square regressions
            cov = np.cov(X.T)
            coef = np.dot(
                np.linalg.pinv(cov[np.ix_(predictors, predictors)]),
                cov[np.ix_(target, predictors)].reshape(predictors.shape[0], 1),
            )
            R = X[:, target] - np.dot(X[:, predictors], coef)
        else:
            self._reg.fit(X[:, predictors], np.ravel(X[:, target]))
            R = X[:, target] - self._reg.predict(X[:, predictors]).reshape(-1, 1)

        return R

    def _fisher_hsic_test(self, X, R, max_p_stat):
        """Conduct statistical test by HSIC and Fisher's method."""
        fisher_stat = 0
        n_features = X.shape[1]

        if n_features == 1:
            fisher_stat, fisher_p = hsic_test_gamma(X, R)
        else:
            for i in range(n_features):
                _, hsic_p = hsic_test_gamma(X[:, [i]], R)
                fisher_stat += np.inf if hsic_p == 0 else -2 * np.log(hsic_p)

                if fisher_stat > max_p_stat:
                    break
            fisher_p = 1 - chi2.cdf(fisher_stat, df=2 * n_features)

        return fisher_p, fisher_stat

    def _flatten(self, arr):
        """Return a copy of an array flattened in one dimension."""
        return [
            val
            for item in arr
            for val in (
                self._flatten(item)
                if hasattr(item, "__iter__") and not isinstance(item, str)
                else [item]
            )
        ]

    def _estimate_adjacency_matrix(self, X, prior_knowledge=None):
        """Estimate adjacency matrix by causal order.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        prior_knowledge : array-like, shape (n_variables, n_variables), optional (default=None)
            Prior knowledge matrix.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if prior_knowledge is not None:
            pk = prior_knowledge.copy()
            np.fill_diagonal(pk, 0)

        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        for i in range(1, len(self._causal_order)):
            target = self._causal_order[i]

            # Flatten the array into one dimension
            predictors = self._flatten(self._causal_order[:i])

            # Exclude variables specified in no_path with prior knowledge
            if prior_knowledge is not None:
                predictors = [p for p in predictors if pk[target, p] != 0]

            # target is exogenous variables if predictors are empty
            if len(predictors) != 0:
                B[target, predictors] = predict_adaptive_lasso(X, predictors, target)

        # Set np.nan if order is unknown
        for unk_order in self._causal_order:
            if hasattr(unk_order, "__iter__") and not isinstance(unk_order, str):
                for i in range(len(unk_order) - 1):
                    xi = unk_order[i]
                    for xj in unk_order[i + 1 :]:
                        B[xi, xj] = np.nan
                        B[xj, xi] = np.nan

        self._adjacency_matrix = B
        return self

    def estimate_total_effect(self, X, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check parameters
        X = check_array(X)

        # Check from/to causal order
        for i, order in enumerate(self._causal_order):
            if hasattr(order, "__iter__") and from_index in order:
                from_order = i
                break
            elif not hasattr(order, "__iter__") and int(from_index) == int(order):
                from_order = i
                break

        for i, order in enumerate(self._causal_order):
            if hasattr(order, "__iter__") and to_index in order:
                to_order = i
                break
            elif not hasattr(order, "__iter__") and int(to_index) == int(order):
                to_order = i
                break

        if from_order > to_order:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_index={to_index}) "
                f"is earlier than the source variable (from_index={from_index})."
            )

        # Check confounders
        if True in np.isnan(self._adjacency_matrix[from_index]):
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the source variable (from_index={from_index}) is influenced by confounders."
            )
            return np.nan

        # from_index + parents indices
        parents = np.where(np.abs(self._adjacency_matrix[from_index]) > 0)[0]
        predictors = [from_index]
        predictors.extend(parents)

        # Estimate total effect
        coefs = predict_adaptive_lasso(X, predictors, to_index)

        return coefs[0]

    def get_error_independence_p_values(self, X):
        """Calculate the p-value matrix of independence between error variables.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        # Check parameters
        X = check_array(X)
        n_samples = X.shape[0]
        n_features = X.shape[1]

        E = X - np.dot(self._adjacency_matrix, X.T).T
        nan_cols = list(set(np.argwhere(np.isnan(self._adjacency_matrix)).ravel()))
        p_values = np.zeros([n_features, n_features])
        for i, j in itertools.combinations(range(n_features), 2):
            if i in nan_cols or j in nan_cols:
                p_values[i, j] = np.nan
                p_values[j, i] = np.nan
            else:
                _, p_value = hsic_test_gamma(
                    np.reshape(E[:, i], [n_samples, 1]),
                    np.reshape(E[:, j], [n_samples, 1]),
                )
                p_values[i, j] = p_value
                p_values[j, i] = p_value

        return p_values

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_features)
            The causal order of fitted model, where
            n_features is the number of features.
            Set the features as a list if order is unknown.
        """
        return self._causal_order

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
            Set np.nan if order is unknown.
        """
        return self._adjacency_matrix

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : BootstrapResult
            Returns the result of bootstrapping.
        """
        # Check parameters
        X = check_array(X)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError("n_sampling must be an integer greater than 0.")
        else:
            raise ValueError("n_sampling must be an integer greater than 0.")

        # Bootstrapping
        adjacency_matrices = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        total_effects = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        for i in range(n_sampling):
            self.fit(resample(X))
            adjacency_matrices[i] = self._adjacency_matrix

            # Calculate total effects
            for c, from_ in enumerate(self._causal_order):
                for to in self._causal_order[c + 1 :]:
                    if hasattr(from_, "__iter__"):
                        for from_item in from_:
                            total_effects[
                                i, to, from_item
                            ] = self.estimate_total_effect(X, from_item, to)
                    else:
                        total_effects[i, to, from_] = self.estimate_total_effect(
                            X, from_, to
                        )

        return BootstrapResult(adjacency_matrices, total_effects)



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""
import graphviz
import numpy as np
from sklearn import linear_model
from sklearn.linear_model import LassoLarsIC, LinearRegression
from sklearn.utils import check_array

import igraph as ig
from scipy.special import expit as sigmoid
import random


__all__ = [
    "print_causal_directions",
    "print_dagc",
    "make_prior_knowledge",
    "remove_effect",
    "make_dot",
    "predict_adaptive_lasso",
    "get_sink_variables",
    "get_exo_variables",
    "find_all_paths",
    "simulate_dag",
    "simulate_parameter",
    "simulate_linear_sem",
    "count_accuracy",
    "set_random_seed",
]


def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)


def simulate_linear_sem(adjacency_matrix, n_samples, sem_type, noise_scale=1.0):
    """Simulate samples from linear SEM with specified type of noise.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_features, n_features)
        Weighted adjacency matrix of DAG, where ``n_features``
        is the number of variables.
    n_samples : int
        Number of samples. n_samples=inf mimics population risk.
    sem_type : str
        SEM type. gauss, exp, gumbel, logistic, poisson.
    noise_scale : float
        scale parameter of additive noise.

    Returns
    -------
    X : array-like, shape (n_samples, n_features)
        Data generated from linear SEM with specified type of noise,
        where ``n_features`` is the number of variables.
    """
    def _simulate_single_equation(X, w):
        """Simulate samples from a single equation.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features_parents)
            Data of parents for a specified variable, where
            n_features_parents is the number of parents.
        w : array-like, shape (1, n_features_parents)
            Weights of parents.

        Returns
        -------
        x : array-like, shape (n_samples, 1)
            Data for the specified variable.
        """
        if sem_type == 'gauss':
            z = np.random.normal(scale=noise_scale, size=n_samples)
            x = X @ w + z
        elif sem_type == 'exp':
            z = np.random.exponential(scale=noise_scale, size=n_samples)
            x = X @ w + z
        elif sem_type == 'gumbel':
            z = np.random.gumbel(scale=noise_scale, size=n_samples)
            x = X @ w + z
        elif sem_type == 'logistic':
            x = np.random.binomial(1, sigmoid(X @ w)) * 1.0
        elif sem_type == 'poisson':
            x = np.random.poisson(np.exp(X @ w)) * 1.0
        elif sem_type == 'subGaussian':
            z = np.random.normal(scale=noise_scale, size=n_samples)
            q = 0.5 + 0.3 * np.random.rand(1)  # sub-Gaussian
            z = np.sign(z) * pow(np.abs(z), q)
            z = z - np.mean(z)
            z = z / np.std(z)
            x = X @ w + z
        elif sem_type == 'supGaussian':
            z = np.random.normal(scale=noise_scale, size=n_samples)
            q = 1.2 + 0.8 * np.random.rand(1)  # super-Gaussian
            z = np.sign(z) * pow(np.abs(z), q)
            z = z - np.mean(z)
            z = z / np.std(z)
            x = X @ w + z
        elif sem_type == 'nonGaussian':
            z = np.random.normal(scale=noise_scale, size=n_samples)
            qq = -1
            if qq == 1:
                q = 0.5 + 0.3 * np.random.rand(1)  # sub-Gaussian
            else:
                q = 1.2 + 0.8 * np.random.rand(1)  # super-Gaussian
            z = np.sign(z) * pow(np.abs(z), q)
            z = z - np.mean(z)
            z = z / np.std(z)
            x = X @ w + z
        elif sem_type == 'uniform':
            z = np.random.uniform(0, 1, n_samples)
            z = z - np.mean(z)
            z = z / np.std(z)
            x = X @ w + z
        elif sem_type == 'gamma':
            z = np.random.gamma(2, 2, n_samples)
            z = z - np.mean(z)
            z = z / np.std(z)
            x = X @ w + z
        elif sem_type == 'laplace':
            z = np.random.laplace(0, scale=noise_scale, size=n_samples)
            x = X @ w + z
        else:
            raise ValueError('unknown sem type')
        return x

    n_features = adjacency_matrix.shape[0]
    if np.isinf(n_samples):
        if sem_type == 'gauss':
            # make 1/n_features X'X = true cov
            X = np.sqrt(n_features) * noise_scale * np.linalg.pinv(np.eye(n_features) - adjacency_matrix)
            return X
        else:
            raise ValueError('population risk not available')
    X = np.zeros([n_samples, n_features])

    G = ig.Graph.Weighted_Adjacency(adjacency_matrix.tolist())
    ordered_vertices = G.topological_sorting()
    assert len(ordered_vertices) == n_features

    for j in ordered_vertices:
        parents = G.neighbors(j, mode=ig.IN)
        X[:, j] = _simulate_single_equation(X[:, parents], adjacency_matrix[parents, j])
    return X


def count_accuracy(W_true, W, W_und=None):
    """Compute recalls and precisions for W, or optionally for CPDAG = W + W_und.

    Parameters
    ----------
    W_true : array-like, shape (n_features, n_features)
        Ground truth graph, where ``n_features`` is
        the number of features.
    W : array-like, shape (n_features, n_features)
        Predicted graph.
    W_und : array-like, shape (n_features, n_features)
        Predicted undirected edges in CPDAG, asymmetric.

    Returns
    -------
    recall : float
        (true positive) / (true positive + false negative).
    precision : float
        (true positive) / (true positive + false positive).
    """
    # convert to binary adjacency matrix
    B_true = (W_true != 0)
    B = (W != 0)
    B_und = None if W_und is None else (W_und != 0)
    # linear index of nonzeros
    pred_und = None
    if B_und is not None:
        pred_und = np.flatnonzero(B_und)
    pred = np.flatnonzero(B)
    cond = np.flatnonzero(B_true)
    cond_reversed = np.flatnonzero(B_true.T)
    cond_skeleton = np.concatenate([cond, cond_reversed])
    # true pos
    true_pos = np.intersect1d(pred, cond, assume_unique=True)
    if B_und is not None:
        # treat undirected edge favorably
        true_pos_und = np.intersect1d(pred_und, cond_skeleton, assume_unique=True)
        true_pos = np.concatenate([true_pos, true_pos_und])
    # false pos
    false_pos = np.setdiff1d(pred, cond_skeleton, assume_unique=True)
    if B_und is not None:
        false_pos_und = np.setdiff1d(pred_und, cond_skeleton, assume_unique=True)
        false_pos = np.concatenate([false_pos, false_pos_und])
    # reverse
    # extra = np.setdiff1d(pred, cond, assume_unique=True)
    # compute ratio
    pred_size = len(pred)
    if B_und is not None:
        pred_size += len(pred_und)
    # fdr = float(len(reverse) + len(false_pos)) / max(pred_size, 1)
    tpr = float(len(true_pos)) / max(len(cond), 1)
    # fpr = float(len(reverse) + len(false_pos)) / max(cond_neg_size, 1)

    recall = tpr
    precision = float(len(true_pos)) / max(pred_size, 1)

    return recall, precision


def simulate_parameter(B, w_ranges=((-2.0, -0.5), (0.5, 2.0))):
    """Simulate SEM parameters for a DAG.

    Parameters
    ----------
    B : array-like, shape (n_features, n_features)
        Binary adjacency matrix of DAG, where ``n_features``
        is the number of features.
    w_ranges : tuple
        Disjoint weight ranges.

    Returns
    -------
    adjacency_matrix : array-like, shape (n_features, n_features)
        Weighted adj matrix of DAG, where ``n_features``
        is the number of features.
    """

    adjacency_matrix = np.zeros(B.shape)
    S = np.random.randint(len(w_ranges), size=B.shape)  # which range
    for i, (low, high) in enumerate(w_ranges):
        U = np.random.uniform(low=low, high=high, size=B.shape)
        adjacency_matrix += B * (S == i) * U
    return adjacency_matrix


def simulate_dag(n_features, n_edges, graph_type):
    """Simulate random DAG with some expected number of edges.

    Parameters
    ----------
    n_features : int
        Number of features.
    n_edges : int
        Expected number of edges.
    graph_type : str
        ER, SF.

    Returns
    -------
    B : array-like, shape (n_features, n_features)
        binary adjacency matrix of DAG.
    """
    def _random_permutation(M):
        # np.random.permutation permutes first axis only
        P = np.random.permutation(np.eye(M.shape[0]))
        return P.T @ M @ P

    def _random_acyclic_orientation(B_und):
        return np.tril(_random_permutation(B_und), k=-1)

    def _graph_to_adjmat(G):
        return np.array(G.get_adjacency().data)

    if graph_type == 'ER':
        # Erdos-Renyi
        G_und = ig.Graph.Erdos_Renyi(n=n_features, m=n_edges)
        B_und = _graph_to_adjmat(G_und)
        B = _random_acyclic_orientation(B_und)
    elif graph_type == 'SF':
        # Scale-free, Barabasi-Albert
        G = ig.Graph.Barabasi(n=n_features, m=int(round(n_edges / n_features)), directed=True)
        B = _graph_to_adjmat(G)
    elif graph_type == 'BP':
        # Bipartite, Sec 4.1 of (Gu, Fu, Zhou, 2018)
        top = int(0.2 * n_features)
        G = ig.Graph.Random_Bipartite(top, n_features - top, m=n_edges, directed=True, neimode=ig.OUT)
        B = _graph_to_adjmat(G)
    else:
        raise ValueError('unknown graph type')
    B_perm = _random_permutation(B)
    assert ig.Graph.Adjacency(B_perm.tolist()).is_dag()
    return B_perm


def print_causal_directions(cdc, n_sampling, labels=None):
    """Print causal directions of bootstrap result to stdout.

    Parameters
    ----------
    cdc : dict
        List of causal directions sorted by count in descending order.
        This can be set the value returned by ``BootstrapResult.get_causal_direction_counts()`` method.
    n_sampling : int
        Number of bootstrapping samples.
    labels : array-like, optional (default=None)
        List of feature lables.
        If set labels, the output feature name will be the specified label.
    """
    for i, (fr, to, co) in enumerate(zip(cdc["from"], cdc["to"], cdc["count"])):
        sign = "" if "sign" not in cdc else "(b>0)" if cdc["sign"][i] > 0 else "(b<0)"
        if labels:
            print(f"{labels[to]} <--- {labels[fr]} {sign} ({100*co/n_sampling:.1f}%)")
        else:
            print(f"x{to} <--- x{fr} {sign} ({100*co/n_sampling:.1f}%)")


def print_dagc(dagc, n_sampling, labels=None):
    """Print DAGs of bootstrap result to stdout.

    Parameters
    ----------
    dagc : dict
        List of directed acyclic graphs sorted by count in descending order.
        This can be set the value returned by ``BootstrapResult.get_directed_acyclic_graph_counts()`` method.
    n_sampling : int
        Number of bootstrapping samples.
    labels : array-like, optional (default=None)
        List of feature lables.
        If set labels, the output feature name will be the specified label.
    """
    for i, (dag, co) in enumerate(zip(dagc["dag"], dagc["count"])):
        print(f"DAG[{i}]: {100*co/n_sampling:.1f}%")
        for j, (fr, to) in enumerate(zip(dag["from"], dag["to"])):
            sign = "" if "sign" not in dag else "(b>0)" if dag["sign"][j] > 0 else "(b<0)"
            if labels:
                print("\t" + f"{labels[to]} <--- {labels[fr]} {sign}")
            else:
                print("\t" + f"x{to} <--- x{fr} {sign}")


def make_prior_knowledge(
    n_variables,
    exogenous_variables=None,
    sink_variables=None,
    paths=None,
    no_paths=None,
):
    """Make matrix of prior knowledge.

    Parameters
    ----------
    n_variables : int
        Number of variables.
    exogenous_variables : array-like, shape (index, ...), optional (default=None)
        List of exogenous variables(index).
        Prior knowledge is created with the specified variables as exogenous variables.
    sink_variables : array-like, shape (index, ...), optional (default=None)
        List of sink variables(index).
        Prior knowledge is created with the specified variables as sink variables.
    paths : array-like, shape ((index, index), ...), optional (default=None)
        List of variables(index) pairs with directed path.
        If ``(i, j)``, prior knowledge is created that xi has a directed path to xj.
    no_paths : array-like, shape ((index, index), ...), optional (default=None)
        List of variables(index) pairs without directed path.
        If ``(i, j)``, prior knowledge is created that xi does not have a directed path to xj.

    Returns
    -------
    prior_knowledge : array-like, shape (n_variables, n_variables)
        Return matrix of prior knowledge used for causal discovery.
    """
    prior_knowledge = np.full((n_variables, n_variables), -1)
    if no_paths:
        for no_path in no_paths:
            prior_knowledge[no_path[1], no_path[0]] = 0
    if paths:
        for path in paths:
            prior_knowledge[path[1], path[0]] = 1
    if sink_variables:
        for var in sink_variables:
            prior_knowledge[:, var] = 0
    if exogenous_variables:
        for var in exogenous_variables:
            prior_knowledge[var, :] = 0
    np.fill_diagonal(prior_knowledge, -1)
    return prior_knowledge


def get_sink_variables(adjacency_matrix):
    """The sink variables(index) in the adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_variables, n_variables)
        Adjacency matrix, where n_variables is the number of variables.

    Returns
    -------
    sink_variables : array-like
        List of sink variables(index).
    """
    am = adjacency_matrix.copy()
    am = np.abs(am)
    np.fill_diagonal(am, 0)
    sink_vars = [i for i in range(am.shape[1]) if am[:, i].sum() == 0]
    return sink_vars


def get_exo_variables(adjacency_matrix):
    """The exogenous variables(index) in the adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_variables, n_variables)
        Adjacency matrix, where n_variables is the number of variables.

    Returns
    -------
    exogenous_variables : array-like
        List of exogenous variables(index).
    """
    am = adjacency_matrix.copy()
    am = np.abs(am)
    np.fill_diagonal(am, 0)
    exo_vars = [i for i in range(am.shape[1]) if am[i, :].sum() == 0]
    return exo_vars


def remove_effect(X, remove_features):
    """Create a dataset that removes the effects of features by linear regression.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    remove_features : array-like
        List of features(index) to remove effects.

    Returns
    -------
    X : array-like, shape (n_samples, n_features)
        Data after removing effects of ``remove_features``.
    """
    X = np.copy(check_array(X))
    features_ = [i for i in np.arange(X.shape[1]) if i not in remove_features]
    for feature in features_:
        reg = linear_model.LinearRegression()
        reg.fit(X[:, remove_features], X[:, feature])
        X[:, feature] = X[:, feature] - reg.predict(X[:, remove_features])
    return X


def make_dot(
    adjacency_matrix,
    labels=None,
    lower_limit=0.01,
    prediction_feature_indices=None,
    prediction_target_label="Y(pred)",
    prediction_line_color="red",
    prediction_coefs=None,
    prediction_feature_importance=None,
    ignore_shape=False,
):
    """Directed graph source code in the DOT language with specified adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like with shape (n_features, n_features)
        Adjacency matrix to make graph, where ``n_features`` is the number of features.
    labels : array-like, optional (default=None)
        Label to use for graph features.
    lower_limit : float, optional (default=0.01)
        Threshold for drawing direction.
        If float, then directions with absolute values of coefficients less than ``lower_limit`` are excluded.
    prediction_feature_indices : array-like, optional (default=None)
        Indices to use as prediction features.
    prediction_target_label : string, optional (default='Y(pred)'))
        Label to use for target variable of prediction.
    prediction_line_color : string, optional (default='red')
        Line color to use for prediction's graph.
    prediction_coefs : array-like, optional (default=None)
        Coefficients to use for prediction's graph.
    prediction_feature_importance : array-like, optional (default=None)
        Feature importance to use for prediction's graph.
    ignore_shape : boolean, optional (default=False)
        Ignore checking the shape of adjaceny_matrix or not.

    Returns
    -------
    graph : graphviz.Digraph
        Directed graph source code in the DOT language.
        If order is unknown, draw a double-headed arrow.
    """
    # Check parameters
    B = check_array(np.nan_to_num(adjacency_matrix))
    if not ignore_shape and B.shape[0] != B.shape[1]:
        raise ValueError("'adjacency_matrix' is not square matrix.")
    if labels is not None:
        if B.shape[1] != len(labels):
            raise ValueError(
                "Length of 'labels' does not match length of 'adjacency_matrix'"
            )
    if prediction_feature_indices is not None:
        if prediction_coefs is not None and (
            len(prediction_feature_indices) != len(prediction_coefs)
        ):
            raise ValueError(
                "Length of 'prediction_coefs' does not match length of 'prediction_feature_indices'"
            )
        if prediction_feature_importance is not None and (
            len(prediction_feature_indices) != len(prediction_feature_importance)
        ):
            raise ValueError(
                "Length of 'prediction_feature_importance' does not match length of 'prediction_feature_indices'"
            )

    d = graphviz.Digraph(engine="dot")

    # nodes
    names = labels if labels else [f"x{i}" for i in range(len(B))]
    for name in names:
        d.node(name)

    # edges
    idx = np.abs(B) > lower_limit
    dirs = np.where(idx)
    for to, from_, coef in zip(dirs[0], dirs[1], B[idx]):
        d.edge(names[from_], names[to], label=f"{coef:.2f}")

    # integrate of prediction model
    if prediction_feature_indices is not None:
        d.node(
            prediction_target_label,
            color=prediction_line_color,
            fontcolor=prediction_line_color,
        )

        if prediction_coefs is not None:
            for from_, coef in zip(prediction_feature_indices, prediction_coefs):
                if np.abs(coef) > lower_limit:
                    d.edge(
                        names[from_],
                        prediction_target_label,
                        label=f"{coef:.2f}",
                        color=prediction_line_color,
                        fontcolor=prediction_line_color,
                        style="dashed",
                    )

        elif prediction_feature_importance is not None:
            for from_, imp in zip(
                prediction_feature_indices, prediction_feature_importance
            ):
                d.edge(
                    names[from_],
                    prediction_target_label,
                    label=f"({imp})",
                    color=prediction_line_color,
                    fontcolor=prediction_line_color,
                    style="dashed",
                )

        else:
            for from_ in prediction_feature_indices:
                d.edge(
                    names[from_],
                    prediction_target_label,
                    color=prediction_line_color,
                    style="dashed",
                )

    # If the value is nan, draw a double-headed arrow
    unk_order = np.where(np.isnan(np.tril(adjacency_matrix)))
    unk_order_set = set([val for item in unk_order for val in item])
    with d.subgraph() as s:
        s.attr(rank="same")
        for node in unk_order_set:
            s.node(names[node])
    for to, from_ in zip(unk_order[0], unk_order[1]):
        d.edge(names[from_], names[to], dir="both")

    return d


def predict_adaptive_lasso(X, predictors, target, gamma=1.0):
    """Predict with Adaptive Lasso.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples
        and n_features is the number of features.
    predictors : array-like, shape (n_predictors)
        Indices of predictor variable.
    target : int
        Index of target variable.

    Returns
    -------
    coef : array-like, shape (n_features)
        Coefficients of predictor variable.
    """
    lr = LinearRegression()
    lr.fit(X[:, predictors], X[:, target])
    weight = np.power(np.abs(lr.coef_), gamma)
    reg = LassoLarsIC(criterion="bic")
    reg.fit(X[:, predictors] * weight, X[:, target])
    return reg.coef_ * weight


def find_all_paths(dag, from_index, to_index, min_causal_effect=0.0):
    """Find all paths from point to point in DAG.

    Parameters
    ----------
    dag : array-like, shape (n_features, n_features)
        The adjacency matrix to fine all paths, where n_features is the number of features.
    from_index : int
        Index of the variable at the start of the path.
    to_index : int
        Index of the variable at the end of the path.
    min_causal_effect : float, optional (default=0.0)
        Threshold for detecting causal direction.
        Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

    Returns
    -------
    paths : array-like, shape (n_paths)
        List of found path, where n_paths is the number of paths.
    effects : array-like, shape (n_paths)
        List of causal effect, where n_paths is the number of paths.
    """
    # Extract all edges
    edges = np.array(np.where(np.abs(np.nan_to_num(dag)) > min_causal_effect)).T

    # Aggregate edges by start point
    to_indices = []
    for i in range(dag.shape[0]):
        adj_list = edges[edges[:, 1] == i][:, 0].tolist()
        if len(adj_list) != 0:
            to_indices.append(adj_list)
        else:
            to_indices.append([])

    # DFS
    paths = []
    stack = [from_index]
    stack_to_indice = [to_indices[from_index]]
    while stack:
        if len(stack) > dag.shape[0]:
            raise ValueError(
                "Unable to find the path because a cyclic graph has been specified."
            )

        cur_index = stack[-1]
        to_indice = stack_to_indice[-1]

        if cur_index == to_index:
            paths.append(stack.copy())
            stack.pop()
            stack_to_indice.pop()
        else:
            if len(to_indice) > 0:
                next_index = to_indice.pop(0)
                stack.append(next_index)
                stack_to_indice.append(to_indices[next_index].copy())
            else:
                stack.pop()
                stack_to_indice.pop()

    # Calculate the causal effect for each path
    effects = []
    for p in paths:
        coefs = [dag[p[i + 1], p[i]] for i in range(len(p) - 1)]
        effects.append(np.cumprod(coefs)[-1])

    return paths, effects



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""

import numpy as np
from scipy.stats import gamma
from statsmodels.nonparametric import bandwidths

__all__ = ["get_kernel_width", "get_gram_matrix", "hsic_teststat", "hsic_test_gamma"]


def get_kernel_width(X):
    """Calculate the bandwidth to median distance between points.
    Use at most 100 points (since median is only a heuristic,
    and 100 points is sufficient for a robust estimate).

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.

    Returns
    -------
    float
        The bandwidth parameter.
    """
    n_samples = X.shape[0]
    if n_samples > 100:
        X_med = X[:100, :]
        n_samples = 100
    else:
        X_med = X

    G = np.sum(X_med * X_med, 1).reshape(n_samples, 1)
    Q = np.tile(G, (1, n_samples))
    R = np.tile(G.T, (n_samples, 1))

    dists = Q + R - 2 * np.dot(X_med, X_med.T)
    dists = dists - np.tril(dists)
    dists = dists.reshape(n_samples ** 2, 1)

    return np.sqrt(0.5 * np.median(dists[dists > 0]))


def _rbf_dot(X, Y, width):
    """Compute the inner product of radial basis functions."""
    n_samples_X = X.shape[0]
    n_samples_Y = Y.shape[0]

    G = np.sum(X * X, 1).reshape(n_samples_X, 1)
    H = np.sum(Y * Y, 1).reshape(n_samples_Y, 1)
    Q = np.tile(G, (1, n_samples_Y))
    R = np.tile(H.T, (n_samples_X, 1))
    H = Q + R - 2 * np.dot(X, Y.T)

    return np.exp(-H / 2 / (width ** 2))


def get_gram_matrix(X, width):
    """Get the centered gram matrices.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.

    width : float
        The bandwidth parameter.

    Returns
    -------
    K, Kc : array
        the centered gram matrices.
    """
    n = X.shape[0]
    H = np.eye(n) - 1 / n * np.ones((n, n))

    K = _rbf_dot(X, X, width)
    Kc = np.dot(np.dot(H, K), H)

    return K, Kc


def hsic_teststat(Kc, Lc, n):
    """get the HSIC statistic.

    Parameters
    ----------
    K, Kc : array
        the centered gram matrices.

    n : float
        the number of samples.

    Returns
    -------
    float
        the HSIC statistic.
    """
    # test statistic m*HSICb under H1
    return 1 / n * np.sum(np.sum(Kc.T * Lc))


def hsic_test_gamma(X, Y, bw_method="mdbs"):
    """get the HSIC statistic.

    Parameters
    ----------
    X, Y : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.

    bw_method : str, optional (default=``mdbs``)
        The method used to calculate the bandwidth of the HSIC.

        * ``mdbs`` : Median distance between samples.
        * ``scott`` : Scott's Rule of Thumb.
        * ``silverman`` : Silverman's Rule of Thumb.

    Returns
    -------
    test_stat : float
        the HSIC statistic.

    p : float
        the HSIC p-value.
    """
    X = X.reshape(-1, 1) if X.ndim == 1 else X
    Y = Y.reshape(-1, 1) if Y.ndim == 1 else Y

    if bw_method == "scott":
        width_x = bandwidths.bw_scott(X)
        width_y = bandwidths.bw_scott(Y)
    elif bw_method == "silverman":
        width_x = bandwidths.bw_silverman(X)
        width_y = bandwidths.bw_silverman(Y)
    # Get kernel width to median distance between points
    else:
        width_x = get_kernel_width(X)
        width_y = get_kernel_width(Y)

    # these are slightly biased estimates of centered gram matrices
    K, Kc = get_gram_matrix(X, width_x)
    L, Lc = get_gram_matrix(Y, width_y)

    # test statistic m*HSICb under H1
    n = X.shape[0]
    bone = np.ones((n, 1))
    test_stat = hsic_teststat(Kc, Lc, n)

    var = (1 / 6 * Kc * Lc) ** 2
    # second subtracted term is bias correction
    var = 1 / n / (n - 1) * (np.sum(np.sum(var)) - np.sum(np.diag(var)))
    # variance under H0
    var = 72 * (n - 4) * (n - 5) / n / (n - 1) / (n - 2) / (n - 3) * var

    K = K - np.diag(np.diag(K))
    L = L - np.diag(np.diag(L))
    mu_X = 1 / n / (n - 1) * np.dot(bone.T, np.dot(K, bone))
    mu_Y = 1 / n / (n - 1) * np.dot(bone.T, np.dot(L, bone))
    # mean under H0
    mean = 1 / n * (1 + mu_X * mu_Y - mu_X - mu_Y)

    alpha = mean ** 2 / var
    # threshold for hsicArr*m
    beta = np.dot(var, n) / mean
    p = 1 - gamma.cdf(test_stat, alpha, scale=beta)[0][0]

    return test_stat, p



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""
import itertools
import warnings

import numpy as np
from sklearn.linear_model import LassoLarsIC, LinearRegression
from sklearn.utils import check_array, resample
from statsmodels.tsa.statespace.varmax import VARMAX

from .base import _BaseLiNGAM
from .bootstrap import BootstrapResult
from .direct_lingam import DirectLiNGAM
from .hsic import hsic_test_gamma
from .utils import predict_adaptive_lasso, find_all_paths


class VARMALiNGAM:
    """Implementation of VARMA-LiNGAM Algorithm [1]_

    References
    ----------
    .. [1] Yoshinobu Kawahara, Shohei Shimizu, Takashi Washio.
       Analyzing relationships among ARMA processes based on non-Gaussianity
       of external influences. Neurocomputing, Volume 74: 2212-2221, 2011
    """

    def __init__(
        self,
        order=(1, 1),
        criterion="bic",
        prune=False,
        max_iter=100,
        ar_coefs=None,
        ma_coefs=None,
        lingam_model=None,
        random_state=None,
    ):
        """Construct a VARMALiNGAM model.

        Parameters
        ----------
        order : turple, length = 2, optional (default=(1, 1))
            Number of lags for AR and MA model.
        criterion : {'aic', 'bic', 'hqic', None}, optional (default='bic')
            Criterion to decide the best order in the all combinations of ``order``.
            Searching the best order is disabled if ``criterion`` is ``None``.
        prune : boolean, optional (default=False)
            Whether to prune the adjacency matrix or not.
        max_iter : int, optional (default=100)
            Maximm number of iterations to estimate VARMA model.
        ar_coefs : array-like, optional (default=None)
            Coefficients of AR of ARMA. Estimating ARMA model is skipped if specified ``ar_coefs`` and `ma_coefs`.
            Shape must be (``order[0]``, n_features, n_features).
        ma_coefs : array-like, optional (default=None)
            Coefficients of MA of ARMA. Estimating ARMA model is skipped if specified ``ar_coefs`` and `ma_coefs`.
            Shape must be (``order[1]``, n_features, n_features).
        lingam_model : lingam object inherits 'lingam._BaseLiNGAM', optional (default=None)
            LiNGAM model for causal discovery. If None, DirectLiNGAM algorithm is selected.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        """
        self._order = order
        self._criterion = criterion
        self._prune = prune
        self._max_iter = max_iter
        self._ar_coefs = (
            check_array(ar_coefs, allow_nd=True) if ar_coefs is not None else None
        )
        self._ma_coefs = (
            check_array(ma_coefs, allow_nd=True) if ma_coefs is not None else None
        )
        self._lingam_model = lingam_model
        self._random_state = random_state

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X: array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        returns
        -------
        self : object
            Returns the instance itself.
        """
        self._causal_order = None
        self._adjacency_matrices = None

        X = check_array(X)

        lingam_model = self._lingam_model
        if lingam_model is None:
            lingam_model = DirectLiNGAM()
        elif not isinstance(lingam_model, _BaseLiNGAM):
            raise ValueError("lingam_model must be a subclass of _BaseLiNGAM")

        phis = self._ar_coefs
        thetas = self._ma_coefs
        order = self._order

        if phis is None or thetas is None:
            phis, thetas, order, residuals = self._estimate_varma_coefs(X)
        else:
            p = phis.shape[0]
            q = thetas.shape[0]
            residuals = self._calc_residuals(X, phis, thetas, p, q)

        model = lingam_model
        model.fit(residuals)

        psis, omegas = self._calc_psi_and_omega(
            model.adjacency_matrix_, phis, thetas, order
        )

        if self._prune:
            ee = np.dot(
                np.eye(model.adjacency_matrix_.shape[0]) - model.adjacency_matrix_,
                residuals.T,
            ).T
            psis, omegas = self._pruning(X, ee, order, model.causal_order_)

        self._ar_coefs = phis
        self._ma_coefs = thetas
        self._order = order
        self._residuals = residuals

        self._causal_order = model.causal_order_
        self._adjacency_matrices = (psis, omegas)

        return self

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : TimeseriesBootstrapResult
            Returns the result of bootstrapping.
        """
        X = check_array(X)

        n_samples = X.shape[0]
        n_features = X.shape[1]
        (p, q) = self._order

        criterion = self._criterion
        self._criterion = None

        self.fit(X)

        residuals = self._residuals
        ar_coefs = self._ar_coefs
        ma_coefs = self._ma_coefs

        total_effects = np.zeros([n_sampling, n_features, n_features * (1 + p)])

        adjacency_matrices = []
        for i in range(n_sampling):
            sampled_residuals = resample(residuals, n_samples=n_samples)

            resampled_X = np.zeros((n_samples, n_features))
            for j in range(n_samples):
                if j < max(p, q):
                    resampled_X[j, :] = sampled_residuals[j]
                    continue

                ar = np.zeros((1, n_features))
                for t, M in enumerate(ar_coefs):
                    ar += np.dot(M, resampled_X[j - t - 1, :].T).T

                ma = np.zeros((1, n_features))
                for t, M in enumerate(ma_coefs):
                    ma += np.dot(M, sampled_residuals[j - t - 1, :].T).T

                resampled_X[j, :] = ar + sampled_residuals[j] + ma

            self.fit(resampled_X)

            psi = self._adjacency_matrices[0]
            omega = self._adjacency_matrices[1]
            am = np.concatenate([*psi, *omega], axis=1)
            adjacency_matrices.append(am)

            ee = np.dot(np.eye(psi[0].shape[0]) - psi[0], sampled_residuals.T).T

            # total effects
            for c, to in enumerate(reversed(self._causal_order)):
                # time t
                for from_ in self._causal_order[: n_features - (c + 1)]:
                    total_effects[i, to, from_] = self.estimate_total_effect(
                        resampled_X, ee, from_, to
                    )

                # time t-tau
                for lag in range(p):
                    for from_ in range(n_features):
                        total_effects[
                            i, to, from_ + n_features
                        ] = self.estimate_total_effect(
                            resampled_X, ee, from_, to, lag + 1
                        )

        self._criterion = criterion

        return VARMABootstrapResult(adjacency_matrices, total_effects, self._order)

    def estimate_total_effect(self, X, E, from_index, to_index, from_lag=0):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        E : array-like, shape (n_samples, n_features)
            Original error data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        X = check_array(X)
        n_features = X.shape[1]

        # Check from/to causal order
        if from_lag == 0:
            from_order = self._causal_order.index(from_index)
            to_order = self._causal_order.index(to_index)
            if from_order > to_order:
                warnings.warn(
                    f"The estimated causal effect may be incorrect because "
                    f"the causal order of the destination variable (to_index={to_index}) "
                    f"is earlier than the source variable (from_index={from_index})."
                )

        # X + lagged X
        X_joined = np.zeros(
            (X.shape[0], X.shape[1] * (1 + from_lag + self._order[0] + self._order[1]))
        )

        for p in range(1 + self._order[0]):
            pos = n_features * p
            X_joined[:, pos : pos + n_features] = np.roll(X[:, 0:n_features], p, axis=0)

        for q in range(self._order[1]):
            pos = n_features * (1 + self._order[0]) + n_features * q
            X_joined[:, pos : pos + n_features] = np.roll(
                E[:, 0:n_features], q + 1, axis=0
            )

        # concat psi and omega
        psi = self._adjacency_matrices[0]
        omega = self._adjacency_matrices[1]
        am = np.concatenate([*psi, *omega], axis=1)

        # from_index + parents indices
        parents = np.where(np.abs(am[from_index]) > 0)[0]
        from_index = from_index if from_lag == 0 else from_index + n_features
        parents = parents if from_lag == 0 else parents + n_features
        predictors = [from_index]
        predictors.extend(parents)

        # Estimate total effect
        coefs = predict_adaptive_lasso(X_joined, predictors, to_index)

        return coefs[0]

    def get_error_independence_p_values(self):
        """Calculate the p-value matrix of independence between error variables.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        eps = self.residuals_
        psi0 = self._adjacency_matrices[0][0]
        E = np.dot(np.eye(psi0.shape[0]) - psi0, eps.T).T
        n_samples = E.shape[0]
        n_features = E.shape[1]

        p_values = np.zeros([n_features, n_features])
        for i, j in itertools.combinations(range(n_features), 2):
            _, p_value = hsic_test_gamma(
                np.reshape(E[:, i], [n_samples, 1]), np.reshape(E[:, j], [n_samples, 1])
            )
            p_values[i, j] = p_value
            p_values[j, i] = p_value

        return p_values

    def _estimate_varma_coefs(self, X):
        if self._criterion not in ["aic", "bic", "hqic"]:
            result = VARMAX(X, order=self._order, trend="c").fit(maxiter=self._max_iter)
        else:
            min_value = float("Inf")
            result = None

            orders = [
                (p, q)
                for p in range(self._order[0] + 1)
                for q in range(self._order[1] + 1)
            ]
            orders.remove((0, 0))

            for order in orders:
                fitted = VARMAX(X, order=order, trend="c").fit(maxiter=self._max_iter)

                value = getattr(fitted, self._criterion)
                if value < min_value:
                    min_value = value
                    result = fitted

        return (
            result.coefficient_matrices_var,
            result.coefficient_matrices_vma,
            result.specification["order"],
            result.resid,
        )

    def _calc_residuals(self, X, ar_coefs, ma_coefs, p, q):
        X = X.T
        n_features = X.shape[0]
        n_samples = X.shape[1]

        start_index = max(p, q)

        epsilon = np.zeros([n_features, n_samples])
        for t in range(n_samples):
            if t < start_index:
                epsilon[:, t] = np.random.normal(size=(n_features))
                continue

            ar = np.zeros([n_features, 1])
            for i in range(p):
                ar += np.dot(ar_coefs[i], X[:, t - i - 1].reshape(-1, 1))

            ma = np.zeros([n_features, 1])
            for j in range(q):
                ma += np.dot(ma_coefs[j], epsilon[:, t - j - 1].reshape(-1, 1))

            epsilon[:, t] = X[:, t] - (ar.flatten() + ma.flatten())

        residuals = epsilon[:, start_index:].T

        return residuals

    def _calc_psi_and_omega(self, psi0, phis, thetas, order):
        psis = [psi0]
        for i in range(order[0]):
            psi = np.dot(np.eye(psi0.shape[0]) - psi0, phis[i])
            psis.append(psi)

        omegas = []
        for j in range(order[1]):
            omega = np.dot(
                np.eye(psi0.shape[0]) - psi0,
                thetas[j],
                np.linalg.inv(np.eye(psi0.shape[0]) - psi0),
            )
            omegas.append(omega)

        return np.array(psis), np.array(omegas)

    def _pruning(self, X, ee, order, causal_order):
        """"""
        n_features = X.shape[1]

        # join X(t), X(t-1) and e(t-1)
        X_joined = np.zeros((X.shape[0], X.shape[1] * (1 + order[0] + order[1])))
        for p in range(1 + order[0]):
            pos = n_features * p
            X_joined[:, pos : pos + n_features] = np.roll(X[:, 0:n_features], p, axis=0)

        for q in range(order[1]):
            pos = n_features * (1 + order[0]) + n_features * q
            X_joined[:, pos : pos + n_features] = np.roll(
                ee[:, 0:n_features], q + 1, axis=0
            )

        # pruned by adaptive lasso
        psi_omega = np.zeros((n_features, n_features * (1 + order[0] + order[1])))
        for i, target in enumerate(causal_order):
            predictors = [
                j for j in range(X_joined.shape[1]) if j not in causal_order[i:]
            ]

            # adaptive lasso
            gamma = 1.0
            lr = LinearRegression()
            lr.fit(X_joined[:, predictors], X_joined[:, target])
            weight = np.power(np.abs(lr.coef_), gamma)
            reg = LassoLarsIC(criterion="bic")
            reg.fit(X_joined[:, predictors] * weight, X_joined[:, target])

            psi_omega[target, predictors] = reg.coef_ * weight

        # split psi and omega
        psis = np.zeros(((1 + order[0]), n_features, n_features))
        for p in range(1 + order[0]):
            pos = n_features * p
            psis[p] = psi_omega[:, pos : pos + n_features]

        omegas = np.zeros((order[1], n_features, n_features))
        for q in range(order[1]):
            pos = n_features * (1 + order[0]) + n_features * q
            omegas[q] = psi_omega[:, pos : pos + n_features]

        return psis, omegas

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_features)
            The causal order of fitted model, where
            n_features is the number of features.
        """
        return self._causal_order

    @property
    def adjacency_matrices_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrices_ : array-like, shape ((p, n_features, n_features), (q, n_features, n_features))
            The adjacency matrix psi and omega of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrices

    @property
    def residuals_(self):
        """Residuals of regression.

        Returns
        -------
        residuals_ : array-like, shape (n_samples)
            Residuals of regression, where n_samples is the number of samples.
        """
        return self._residuals


class VARMABootstrapResult(BootstrapResult):
    """The result of bootstrapping for Time series algorithm."""

    def __init__(self, adjacency_matrices, total_effects, order):
        """Construct a BootstrapResult.

        Parameters
        ----------
        adjacency_matrices : array-like, shape (n_sampling)
            The adjacency matrix list by bootstrapping.
        total_effects : array-like, shape (n_sampling)
            The total effects list by bootstrapping.
        """
        self._order = order
        super().__init__(adjacency_matrices, total_effects)

    def get_paths(self, from_index, to_index, from_lag=0, to_lag=0, min_causal_effect=None):
        """Get all paths from the start variable to the end variable and their bootstrap probabilities.

        Parameters
        ----------
        from_index : int
            Index of the variable at the start of the path.
        to_index : int
            Index of the variable at the end of the path.
        from_lag : int
            Number of lag at the start of the path.
            ``from_lag`` should be greater than or equal to ``to_lag``.
        to_lag : int
            Number of lag at the end of the path.
            ``from_lag`` should be greater than or equal to ``to_lag``.
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        paths : dict
            List of path and bootstrap probability.
            The dictionary has the following format::

            {'path': [n_paths], 'effect': [n_paths], 'probability': [n_paths]}

            where ``n_paths`` is the number of paths.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")
        if to_lag > from_lag:
            raise ValueError("from_lag should be greater than or equal to to_lag.")
        if to_lag == from_lag:
            if to_index == from_index:
                raise ValueError("The same variable is specified for from and to.")

        # Find all paths from from_index to to_index
        paths_list = []
        effects_list = []
        for am in self._adjacency_matrices:
            n_features = am.shape[0]
            expansion_m_size = n_features * (self._order[0] + 1)
            expansion_m = np.zeros((expansion_m_size, expansion_m_size))
            n_lags = int(expansion_m_size / n_features) - 1
            for i in range(n_lags + 1):
                for j in range(i, n_lags + 1):
                    row = n_features * i
                    col = n_features * j
                    lag = col - row
                    expansion_m[row:row + n_features, col:col + n_features] = am[0:n_features, lag:lag + n_features]
            paths, effects = find_all_paths(expansion_m,
                                            int(n_features * from_lag + from_index),
                                            int(n_features * to_lag + to_index),
                                            min_causal_effect)

            # Convert path to string to make them easier to handle.
            paths_list.extend(["_".join(map(str, p)) for p in paths])
            effects_list.extend(effects)

        paths_list = np.array(paths_list)
        effects_list = np.array(effects_list)

        # Count paths
        paths_str, counts = np.unique(paths_list, axis=0, return_counts=True)

        # Sort by count
        order = np.argsort(-counts)
        probs = counts[order] / len(self._adjacency_matrices)
        paths_str = paths_str[order]

        # Calculate median of causal effect for each path
        effects = [
            np.median(effects_list[np.where(paths_list == p)]) for p in paths_str
        ]

        result = {
            "path": [[int(i) for i in p.split("_")] for p in paths_str],
            "effect": effects,
            "probability": probs.tolist(),
        }
        return result



"""
The lingam module includes implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""

from .bootstrap import BootstrapResult
from .bottom_up_parce_lingam import BottomUpParceLiNGAM
from .causal_effect import CausalEffect
from .direct_lingam import DirectLiNGAM
from .ica_lingam import ICALiNGAM
from .longitudinal_lingam import LongitudinalLiNGAM, LongitudinalBootstrapResult
from .multi_group_direct_lingam import MultiGroupDirectLiNGAM
from .rcd import RCD
from .var_lingam import VARLiNGAM, VARBootstrapResult
from .varma_lingam import VARMALiNGAM, VARMABootstrapResult
from .lina import LiNA
from .lina import MDLiNA
from .resit import RESIT
__all__ = [
    "ICALiNGAM",
    "DirectLiNGAM",
    "BootstrapResult",
    "MultiGroupDirectLiNGAM",
    "CausalEffect",
    "VARLiNGAM",
    "VARMALiNGAM",
    "LongitudinalLiNGAM",
    "VARBootstrapResult",
    "VARMABootstrapResult",
    "LongitudinalBootstrapResult",
    "BottomUpParceLiNGAM",
    "RCD",
    "LiNA",
    "MDLiNA",
    "RESIT",
]

__version__ = "1.6.0"



import os

import numpy as np
import pandas as pd

from lingam.base import _BaseLiNGAM
from lingam.causal_effect import CausalEffect

class DummyLiNGAM(_BaseLiNGAM):
    def __init__(self, random_state=None, causal_order=None, adjacency_matrix=None):
        super().__init__(random_state)
        self._causal_order = causal_order
        self._adjacency_matrix = adjacency_matrix
    def fit(self, X):
        pass

class DummyPrediction(object):
    def __init__(self, reval_predict=None, coef=None, intercept=None):
        self._reval_predict = reval_predict
        self._coef = coef
        self._intercept = intercept
    def predict(self, X):
        return self._reval_predict
    @property
    def coef_(self):
        return self._coef
    @property
    def intercept_(self):
        return self._intercept

class DummyPrediction2(object):
    def __init__(self, reval_predict=None, coef=None, intercept=None):
        self._reval_predict = reval_predict
        self._coef = coef
        self._intercept = intercept
    def predict_proba(self, X):
        return self._reval_predict
    @property
    def coef_(self):
        return self._coef
    @property
    def intercept_(self):
        return self._intercept

def test_estimate_effects_on_prediction_success():
    # causal direction: x0 --> x1, x0 --> x2, x1 --> x2
    x0 = np.random.uniform(size=1000)
    x1 = -1.0*x0 + np.random.uniform(size=1000)
    x2 = 3.0*x0 + 2.0*x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2]).T, columns=['x0', 'x1', 'x2'])

    # specified LiNGAM for causal model
    model = DummyLiNGAM(causal_order=[0, 1, 2],
        adjacency_matrix=np.array([
            [ 0., 0., 0.],
            [-1., 0., 0.],
            [ 3., 2., 0.],
        ]))
    ce = CausalEffect(model)
    pred_model = DummyPrediction(reval_predict=np.array([0]))
    effects = ce.estimate_effects_on_prediction(X, 2, pred_model)
    assert effects.shape == (3, 2)

    # retry
    effects = ce.estimate_effects_on_prediction(X, 2, pred_model)

    En = np.array([0.1, 0.1, 0.1])
    effects = ce._get_propagated_effects(En, 0, 1)
    assert effects[0] == 1
    assert effects[1] == -0.9
    assert effects[2] == (3.0+(-1.8)+0.1)

    # specified np.array for causal model
    ce = CausalEffect(np.array([[0., -1., 0.], [ 0., 0., 0.], [ 3., 2., 0.]]))
    pred_model = DummyPrediction(reval_predict=np.array([0]))
    effects = ce.estimate_effects_on_prediction(X, 2, pred_model)
    assert effects.shape == (3, 2)

    # specified list for causal model
    ce = CausalEffect([[0., -1., 0.], [ 0., 0., 0.], [ 3., 2., 0.]])
    pred_model = DummyPrediction(reval_predict=np.array([0]))
    effects = ce.estimate_effects_on_prediction(X, 2, pred_model)
    assert effects.shape == (3, 2)

    # predict_proba
    ce = CausalEffect([[0., -1., 0.], [ 0., 0., 0.], [ 3., 2., 0.]])
    pred_model = DummyPrediction2(reval_predict=np.array([[0, 0]]))
    effects = ce.estimate_effects_on_prediction(X, 2, pred_model)
    assert effects.shape == (3, 2)

def test_estimate_effects_on_prediction_invalid_input():
    # causal direction: x0 --> x1, x0 --> x2, x1 --> x2
    x0 = np.random.uniform(size=1000)
    x1 = -1.0*x0 + np.random.uniform(size=1000)
    x2 = 3.0*x0 + 2.0*x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2]).T, columns=['x0', 'x1', 'x2'])

    # Not causal model
    try:
        ce = CausalEffect(1)
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not matrix
    try:
        ce = CausalEffect(np.array([0, 1, 2]))
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ce = CausalEffect([0, 1, 2])
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not square matrix
    try:
        ce = CausalEffect(np.array([[0, 1, 2], [0, 0, 0]]))
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ce = CausalEffect([[0, 1, 2], [0, 0, 0]])
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not triangular matrix
    try:
        ce = CausalEffect(np.array([[0, 1, 2], [1, 1, 1], [2, 3, 4]]))
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ce = CausalEffect([[0, 1, 2], [1, 1, 1], [2, 3, 4]])
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not prediction model
    try:
        ce = CausalEffect(np.array([[0, 0, 0], [1, 0, 0], [2, 2, 0]]))
        pred_model = [0]
        ce.estimate_effects_on_prediction(X, 2, pred_model)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_estimate_optimal_intervention_success():
    # causal direction: x0 --> x1, x0 --> x2, x1 --> x2
    x0 = np.random.uniform(size=1000)
    x1 = -1.0*x0 + np.random.uniform(size=1000)
    x2 = 3.0*x0 + 2.0*x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2]).T, columns=['x0', 'x1', 'x2'])

    # specified LiNGAM for causal model
    model = DummyLiNGAM(causal_order=[0, 1, 2],
        adjacency_matrix=np.array([
            [ 0., 0., 0.],
            [-1., 0., 0.],
            [ 3., 2., 0.],
        ]))
    ce = CausalEffect(model)
    pred_model = DummyPrediction(reval_predict=np.array([1]), coef=np.array([1, 2]), intercept=0)
    optimal_intervention = ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    assert optimal_intervention > 0.95 and optimal_intervention < 1.05

    # specified np.array for causal model
    ce = CausalEffect(np.array([[ 0., 0., 0.], [-1., 0., 0.], [ 3., 2., 0.]]))
    pred_model = DummyPrediction(reval_predict=np.array([1]), coef=np.array([1, 2]), intercept=0)
    optimal_intervention = ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    assert optimal_intervention > 0.95 and optimal_intervention < 1.05

    # specified list for causal model
    ce = CausalEffect([[ 0., 0., 0.], [-1., 0., 0.], [ 3., 2., 0.]])
    pred_model = DummyPrediction(reval_predict=np.array([1]), coef=np.array([1, 2]), intercept=0)
    optimal_intervention = ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    assert optimal_intervention > 0.95 and optimal_intervention < 1.05

def test_estimate_optimal_intervention_invalid_input():
    # causal direction: x0 --> x1, x0 --> x2, x1 --> x2
    x0 = np.random.uniform(size=1000)
    x1 = -1.0*x0 + np.random.uniform(size=1000)
    x2 = 3.0*x0 + 2.0*x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2]).T, columns=['x0', 'x1', 'x2'])

    # Not causal model
    try:
        ce = CausalEffect(1)
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not matrix
    try:
        ce = CausalEffect(np.array([0, 1, 2]))
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ce = CausalEffect([0, 1, 2])
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not square matrix
    try:
        ce = CausalEffect(np.array([[0, 1, 2], [0, 0, 0]]))
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ce = CausalEffect([[0, 1, 2], [0, 0, 0]])
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not triangular matrix
    try:
        ce = CausalEffect(np.array([[0, 1, 2], [1, 1, 1], [2, 3, 4]]))
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ce = CausalEffect([[0, 1, 2], [1, 1, 1], [2, 3, 4]])
        pred_model = DummyPrediction(reval_predict=np.array([0]))
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not prediction model
    try:
        ce = CausalEffect(np.array([[0, 0, 0], [1, 0, 0], [2, 2, 0]]))
        pred_model = [0]
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # predict_proba
    try:
        ce = CausalEffect([[ 0., 0., 0.], [-1., 0., 0.], [ 3., 2., 0.]])
        pred_model = DummyPrediction2(reval_predict=np.array([[1, 1]]), coef=np.array([1, 2]), intercept=0)
        ce.estimate_optimal_intervention(X, 2, pred_model, 0, 0)
    except ValueError:
        pass
    else:
        raise AssertionError
        




import os

import numpy as np
import pandas as pd
from lingam.bootstrap import BootstrapMixin, BootstrapResult


class DummyBoostrapMixin(BootstrapMixin):
    def __init__(self, random_state=None):
        self._random_state = random_state
        self._causal_order = [0, 1, 2, 3]
        self._adjacency_matrix = [
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
        ]
        self._adjacency_matrices = [
            # causal direction: x0 --> x1 --> x2 --> x3
            [
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
            ],
            # causal direction: x1 --> x2 --> x3
            # However, a small causal effect is included
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.1, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
            ],
            # causal direction: x2 --> x3
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
            ],
            # causal direction: None
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            # causal direction: x0 --> x1 --> x2 --> x3
            # However, a negative causal effect is included
            [
                [0.0, 0.0, 0.0, 0.0],
                [2.0, 0.0, 0.0, 0.0],
                [0.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, -0.5, 0.0],
            ],
        ]

    def fit(self, X):
        self._adjacency_matrix = np.array(self._adjacency_matrices.pop(0))
        return self

    def estimate_total_effect(self, X, from_index, to_index):
        table = np.array([
            [0.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0],
            [2.0, 1.0, 0.0, 0.0],
            [3.0, 2.0, 1.0, 0.0],
        ])
        return table[to_index, from_index]

    @property
    def causal_order_(self):
        return self._causal_order

    @property
    def adjacency_matrix_(self):
        return self._adjacency_matrix


class DummyBoostrapMixinNonCausal(BootstrapMixin):
    def __init__(self, random_state=None):
        self._random_state = random_state
        self._causal_order = [0, 1]
        self._adjacency_matrix = [
                [0.0, 0.0],
                [0.0, 0.0],
        ]
        self._adjacency_matrices = [
            # causal direction: None
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
            # causal direction: None
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
            # causal direction: None
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
            # causal direction: None
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
            # causal direction: None
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
        ]

    def estimate_total_effect(self, X, from_index, to_index):
        table = np.array([
            [0.0, 0.0],
            [0.0, 0.0],
        ])
        return table[to_index, from_index]

    def fit(self, X):
        self._adjacency_matrix = np.array(self._adjacency_matrices.pop(0))
        return self

    @property
    def causal_order_(self):
        return self._causal_order

    @property
    def adjacency_matrix_(self):
        return self._adjacency_matrix


def test_bootstrap_success():
    x0 = np.random.uniform(size=1000)
    x1 = np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                     columns=['x0', 'x1', 'x2', 'x3'])

    model = DummyBoostrapMixin()
    result = model.bootstrap(X, n_sampling=5)
    am = result.adjacency_matrices_
    assert len(am) == 5
    te = result.total_effects_
    assert len(te) == 5

    # No argument
    cdc = result.get_causal_direction_counts()
    assert cdc['from'] == [2, 1, 0]
    assert cdc['to'] == [3, 2, 1]
    assert cdc['count'] == [4, 3, 2]

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)
    assert cdc['from'] == [2, 1]
    assert cdc['to'] == [3, 2]
    assert cdc['count'] == [4, 3]

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)
    assert cdc['from'] == [2, 0, 1]
    assert cdc['to'] == [3, 1, 2]
    assert cdc['count'] == [4, 2, 2]

    # min_causal_effect=0.6
    cdc = result.get_causal_direction_counts(min_causal_effect=0.6)
    assert cdc['from'] == [2, 0, 1]
    assert cdc['to'] == [3, 1, 2]
    assert cdc['count'] == [3, 2, 2]

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)
    assert cdc['from'] == [1, 2, 0, 2]
    assert cdc['to'] == [2, 3, 1, 3]
    assert cdc['count'] == [3, 3, 2, 1]
    assert cdc['sign'] == [1, 1, 1, -1]

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()
    assert len(dagc['dag']) == 4
    assert dagc['dag'][0]['from'] == [0, 1, 2]
    assert dagc['dag'][0]['to'] == [1, 2, 3]
    assert dagc['count'][0] == 2
    assert not dagc['dag'][1]['from']
    assert not dagc['dag'][1]['to']
    assert dagc['count'][1] == 1
    assert dagc['dag'][2]['from'] == [2]
    assert dagc['dag'][2]['to'] == [3]
    assert dagc['count'][2] == 1
    assert dagc['dag'][3]['from'] == [1, 2]
    assert dagc['dag'][3]['to'] == [2, 3]
    assert dagc['count'][3] == 1

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)
    assert len(dagc['dag']) == 2
    assert dagc['dag'][0]['from'] == [0, 1, 2]
    assert dagc['dag'][0]['to'] == [1, 2, 3]
    assert dagc['count'][0] == 2
    assert not dagc['dag'][1]['from']
    assert not dagc['dag'][1]['to']
    assert dagc['count'][1] == 1

    # min_causal_effect=0.2
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.2)
    assert len(dagc['dag']) == 3
    assert dagc['dag'][0]['from'] == [2]
    assert dagc['dag'][0]['to'] == [3]
    assert dagc['count'][0] == 2
    assert dagc['dag'][1]['from'] == [0, 1, 2]
    assert dagc['dag'][1]['to'] == [1, 2, 3]
    assert dagc['count'][1] == 2
    assert not dagc['dag'][2]['from']
    assert not dagc['dag'][2]['to']
    assert dagc['count'][2] == 1

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)
    assert len(dagc['dag']) == 4
    assert dagc['dag'][0]['from'] == [2]
    assert dagc['dag'][0]['to'] == [3]
    assert dagc['count'][0] == 2
    assert not dagc['dag'][1]['from']
    assert not dagc['dag'][1]['to']
    assert dagc['count'][1] == 1
    assert dagc['dag'][2]['from'] == [0, 1]
    assert dagc['dag'][2]['to'] == [1, 2]
    assert dagc['count'][2] == 1
    assert dagc['dag'][3]['from'] == [0, 1, 2]
    assert dagc['dag'][3]['to'] == [1, 2, 3]
    assert dagc['count'][3] == 1

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)
    assert len(dagc['dag']) == 5
    print(dagc)
    assert not dagc['dag'][0]['from']
    assert not dagc['dag'][0]['to']
    assert not dagc['dag'][0]['sign']
    assert dagc['count'][0] == 1

    assert dagc['dag'][1]['from'] == [2]
    assert dagc['dag'][1]['to'] == [3]
    assert dagc['dag'][1]['sign'] == [1]
    assert dagc['count'][1] == 1

    assert dagc['dag'][2]['from'] == [1, 2]
    assert dagc['dag'][2]['to'] == [2, 3]
    assert dagc['dag'][2]['sign'] == [1, 1]
    assert dagc['count'][2] == 1

    assert dagc['dag'][3]['from'] == [0, 1, 2]
    assert dagc['dag'][3]['to'] == [1, 2, 3]
    assert dagc['dag'][3]['sign'] == [1, 1, -1]
    assert dagc['count'][3] == 1

    assert dagc['dag'][4]['from'] == [0, 1, 2]
    assert dagc['dag'][4]['to'] == [1, 2, 3]
    assert dagc['dag'][4]['sign'] == [1, 1, 1]
    assert dagc['count'][4] == 1

    # get_probabilities
    probs = result.get_probabilities()
    assert probs.shape[0]==4 and probs.shape[1]==4

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)
    assert probs.shape[0]==4 and probs.shape[1]==4

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_path
    paths = result.get_paths(0, 1)

    # get_path
    paths = result.get_paths(0, 1, min_causal_effect=0.6)


def test_bootstrap_invalid_data():
    # Not array data
    X = 1
    try:
        model = DummyBoostrapMixin()
        result = model.bootstrap(X, n_sampling=5)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = DummyBoostrapMixin()
        result = model.bootstrap(X, n_sampling=5)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.nan
    try:
        model = DummyBoostrapMixin()
        result = model.bootstrap(X, n_sampling=5)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf
    try:
        model = DummyBoostrapMixin()
        result = model.bootstrap(X, n_sampling=5)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Non-causal data
    x0 = np.random.uniform(size=1)
    x1 = np.random.uniform(size=1)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    model = DummyBoostrapMixinNonCausal()
    result = model.bootstrap(X, n_sampling=5)

    cdc = result.get_causal_direction_counts()
    assert not cdc['from'] and not cdc['to'] and not cdc['count']

    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)
    assert not cdc['from'] and not cdc['to'] and not cdc['count'] and not cdc['sign']

    dagc = result.get_directed_acyclic_graph_counts()
    assert not dagc['dag'][0]['from'] and not dagc['dag'][0]['to'] and dagc['count'][0] == 5

    x0 = np.random.uniform(size=1000)
    x1 = np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = np.random.uniform(size=1000)
    X_success = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                             columns=['x0', 'x1', 'x2', 'x3'])

    # Invalid argument: bootstrap(n_sampling=-1)
    model = DummyBoostrapMixin()
    try:
        result = model.bootstrap(X_success, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: bootstrap(n_sampling='3')
    model = DummyBoostrapMixin()
    try:
        result = model.bootstrap(X_success, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_causal_direction_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_directed_acyclic_graph_counts(n_dags=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_directed_acyclic_graph_counts(n_dags='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_probabilities(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_total_causal_effects(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(min_causal_effect=-1.0)
    model = DummyBoostrapMixin()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_paths(0, 1, min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError




"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""


import numpy as np
import scipy.linalg
import scipy.optimize as sopt
from scipy.special import expit as sigmoid
from factor_analyzer import (ConfirmatoryFactorAnalyzer, ModelSpecificationParser)
import random
from statsmodels.stats.outliers_influence import variance_inflation_factor
import pandas as pd
import lingam
import lingam.utils as ut


def test_fit_lina():
    i = np.random.randint(1000)
    ut.set_random_seed(i)
    noise_ratios = 0.1
    n_features = 10
    n_samples, n_features_latent, n_edges, graph_type, sem_type = 1000, 5, 5, 'ER', 'laplace'
    B_true = ut.simulate_dag(n_features_latent, n_edges, graph_type)
    W_true = ut.simulate_parameter(B_true)  # row to column

    f = ut.simulate_linear_sem(W_true, n_samples, sem_type)
    f_nor = np.zeros([n_samples, n_features_latent])
    scale = np.zeros([1, n_features_latent])
    W_true_scale = np.zeros([n_features_latent, n_features_latent])
    for j in range(n_features_latent):
        scale[0, j] = np.std(f[:, j])
        f_nor[:, j] = f[:, j] / np.std(f[:, j])
        W_true_scale[:, j] = W_true[:, j] / scale[0, j]  # scaled W_true

    # generate noises ei of xi
    e = np.random.random([n_features, n_samples])
    for j in range(n_features):
        e[j, :] = e[j, :] - np.mean(e[j, :])
        e[j, :] = e[j, :] / np.std(e[j, :])

    G = np.zeros([n_features, n_features_latent])
    G[0, 0] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[1, 0] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[2, 1] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[3, 1] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[4, 2] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[5, 2] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[6, 3] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[7, 3] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[8, 4] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G[9, 4] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G_sign = np.sign(G)

    # normalize G
    G_nor = np.zeros([n_features, n_features_latent])
    for j in range(n_features):
        e[j, :] = e[j, :] / np.sqrt(np.square(np.sum(G[j, :])) + np.square(noise_ratios))
        G_nor[j, :] = G[j, :] / np.sqrt(np.square(np.sum(G[j, :])) + np.square(noise_ratios))

    X = G_nor @ f_nor.T + noise_ratios * e  # "e is small or n_features are large"
    X = X.T
    print(G_nor)

    model = lingam.LiNA()
    model.fit(X, G_sign, scale)

    print('The estimated adjacency matrix is:\n', model.adjacency_matrix_)
    print('The true adjacency matrix is:\n', W_true)


def test_fit_mdlina():
    i = np.random.randint(1000)
    ut.set_random_seed(i)
    n_features = 6
    noise_ratios = 0.1

    n_samples, n_features_latent, n_edges, graph_type, sem_type1, sem_type2 = \
        1000, 3, 3, 'ER', 'subGaussian', 'supGaussian'
    B_true = ut.simulate_dag(n_features_latent, n_edges, graph_type)  # skeleton btw. latent factors
    W_true = ut.simulate_parameter(B_true)  # causal effects matrix btw. latent factors

    # 1 domain
    f = ut.simulate_linear_sem(W_true, n_samples, sem_type1)
    f_nor1 = np.zeros([n_samples, n_features_latent])
    scale1 = np.zeros([1, n_features_latent])
    W_true_scale = np.zeros([n_features_latent, n_features_latent])
    for j in range(n_features_latent):
        scale1[0, j] = np.std(f[:, j])
        f_nor1[:, j] = f[:, j] / np.std(f[:, j])
        W_true_scale[:, j] = W_true[:, j] / scale1[0, j]
    e = np.random.random([n_features, n_samples])
    for j in range(n_features):
        e[j, :] = e[j, :] - np.mean(e[j, :])
        e[j, :] = e[j, :] / np.std(e[j, :])

    G1 = np.zeros([n_features, n_features_latent])
    G1[0, 0] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G1[1, 0] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G1[2, 1] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G1[3, 1] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G1[4, 2] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G1[5, 2] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G_sign1 = np.sign(G1)
    # normalize G
    G_nor1 = np.zeros([n_features, n_features_latent])
    for j in range(n_features):
        e[j, :] = e[j, :] / np.sqrt(np.square(np.sum(G1[j, :])) + np.square(noise_ratios))
        G_nor1[j, :] = G1[j, :] / np.sqrt(np.square(np.sum(G1[j, :])) + np.square(noise_ratios))
    X1 = G_nor1 @ f_nor1.T + noise_ratios * e  # "the noise ratio e is small or n_features is large"
    X1 = X1.T

    # 2 domain
    f2 = ut.simulate_linear_sem(W_true, n_samples, sem_type2)
    f_nor2 = np.zeros([n_samples, n_features_latent])
    scale2 = np.zeros([1, n_features_latent])
    W_true_scale = np.zeros([n_features_latent, n_features_latent])
    for j in range(n_features_latent):
        scale2[0, j] = np.std(f2[:, j])
        f_nor2[:, j] = f2[:, j] / np.std(f2[:, j])
        W_true_scale[:, j] = W_true[:, j] / scale2[0, j]
    e = np.random.random([n_features, n_samples])
    for j in range(n_features):
        e[j, :] = e[j, :] - np.mean(e[j, :])
        e[j, :] = e[j, :] / np.std(e[j, :])
    G2 = np.zeros([n_features, n_features_latent])
    G2[0, 0] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G2[1, 0] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G2[2, 1] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G2[3, 1] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G2[4, 2] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G2[5, 2] = random.choice((-1, 1)) * (0.2 + 0.5 * np.random.rand(1))
    G_sign2 = np.sign(G2)
    # normalize G
    G_nor2 = np.zeros([n_features, n_features_latent])
    for j in range(n_features):
        e[j, :] = e[j, :] / np.sqrt(np.square(np.sum(G2[j, :])) + np.square(noise_ratios))
        G_nor2[j, :] = G2[j, :] / np.sqrt(np.square(np.sum(G2[j, :])) + np.square(noise_ratios))
    X2 = G_nor2 @ f_nor2.T + noise_ratios * e
    X2 = X2.T

    # augment the data X
    XX = scipy.linalg.block_diag(X1, X2)
    G_sign = scipy.linalg.block_diag(G_sign1, G_sign2)
    scale = scipy.linalg.block_diag(scale1, scale2)
    print(G_sign)

    model = lingam.MDLiNA()
    model.fit(XX, G_sign, scale)

    print('The estimated adjacency matrix is:\n', model._adjacency_matrix)
    print('The true adjacency matrix is:\n', W_true)
